from packageurl import PackageURL
from pydantic import ValidationError

from labels.model.ecosystem_data.python import WheelEggEcosystemData
from labels.model.file import Location
from labels.model.package import Ecosystem, Language, Package
from labels.parsers.cataloger.utils import log_malformed_package_warning
from labels.utils.strings import normalize_name


def new_python_package(
    *,
    name: str | None,
    version: str | None,
    location: Location,
    ecosystem_data: WheelEggEcosystemData | None = None,
) -> Package | None:
    if not name or not version:
        return None

    normalized_name = normalize_name(name, Ecosystem.PYPI)

    p_url = PackageURL(type="pypi", name=normalized_name, version=version).to_string()

    try:
        return Package(
            name=normalized_name,
            version=version,
            locations=[location],
            language=Language.PYTHON,
            ecosystem=Ecosystem.PYPI,
            ecosystem_data=ecosystem_data,
            p_url=p_url,
        )
    except ValidationError as ex:
        log_malformed_package_warning(location, ex)
        return None



import logging
import os
from typing import NamedTuple

import boto3

from labels.model.core import SourceType

LOGGER = logging.getLogger(__name__)


class ExecutionMetrics(NamedTuple):
    """Metrics collected during execution.

    Attributes:
       group: Group identifier for the metrics
       technique_time: Execution time in seconds

    """

    group: str
    technique_time: float


def is_fluid_batch_env() -> bool:
    return "AWS_BATCH_JOB_ID" in os.environ


def process_sbom_metrics(
    execution_id: str | None,
    technique_time: float,
    source_type: SourceType,
) -> None:
    if source_type != SourceType.DIRECTORY:
        return
    if not is_fluid_batch_env():
        return
    try:
        if (
            execution_id
            and (execution_id_parts := execution_id.split("_"))
            and len(execution_id_parts) >= 3
        ):
            metrics = ExecutionMetrics(
                group=execution_id_parts[0],
                technique_time=round(technique_time, 2),
            )
            send_metrics_to_cloudwatch(metrics)
        else:
            LOGGER.error(
                "Unable to send metrics to cloudwatch",
                extra={
                    "extra": {
                        "execution_id": execution_id,
                    }
                },
            )
    except Exception:
        LOGGER.exception("Error sending metrics to cloudwatch")


def send_metrics_to_cloudwatch(execution_metrics: ExecutionMetrics) -> None:
    try:
        dimensions = [
            {"Name": "Group", "Value": execution_metrics.group},
        ]

        metric_data = [
            {
                "MetricName": "ExecutionTime",
                "Dimensions": dimensions,
                "Value": execution_metrics.technique_time,
                "Unit": "Seconds",
            },
        ]

        LOGGER.debug(
            "Sending metrics to CloudWatch",
            extra={
                "extra": {
                    "namespace": "LabelsMetrics",
                    "group": execution_metrics.group,
                    "execution_time": execution_metrics.technique_time,
                }
            },
        )
        cloudwatch_client = boto3.client("cloudwatch", "us-east-1")  # type: ignore[misc]
        cloudwatch_client.put_metric_data(  # type: ignore[misc]
            Namespace="LabelsMetrics",
            MetricData=metric_data,
        )
        LOGGER.info("CloudWatch metrics sent successfully")
    except Exception as exc:  # noqa: BLE001
        LOGGER.warning("Unable to send metrics to cloudwatch: %s", exc)



from fnmatch import fnmatch

import reactivex
from reactivex.abc import ObserverBase, SchedulerBase

from labels.model.parser import Request
from labels.parsers.cataloger.javascript.parse_bun_lock import parse_bun_lock
from labels.parsers.cataloger.javascript.parse_html_scripts import parse_html_scripts
from labels.parsers.cataloger.javascript.parse_package_json import parse_package_json
from labels.parsers.cataloger.javascript.parse_package_lock.package_lock_dispatcher import (
    parse_package_lock,
)
from labels.parsers.cataloger.javascript.parse_pnpm_lock import parse_pnpm_lock
from labels.parsers.cataloger.javascript.parse_yarn_lock import parse_yarn_lock


def on_next_javascript(
    source: reactivex.Observable[str],
) -> reactivex.Observable[Request]:
    def subscribe(
        observer: ObserverBase[Request],
        scheduler: SchedulerBase | None = None,
    ) -> reactivex.abc.DisposableBase:
        def on_next(value: str) -> None:
            patterns = [
                (
                    ("**/package.json", "package.json"),
                    parse_package_json,
                    "javascript-parse-package-json",
                ),
                (
                    ("**/package-lock.json", "package-lock.json"),
                    parse_package_lock,
                    "javascript-parse-package-lock",
                ),
                (
                    ("**/yarn.lock", "yarn.lock"),
                    parse_yarn_lock,
                    "javascript-parse-yarn-lock",
                ),
                (
                    ("**/pnpm-lock.yaml", "pnpm-lock.yaml"),
                    parse_pnpm_lock,
                    "javascript-parse-pnpm-lock",
                ),
                (
                    ("**/bun.lock", "bun.lock"),
                    parse_bun_lock,
                    "javascript-parse-bun-lock",
                ),
                (
                    ("*.html",),
                    parse_html_scripts,
                    "javascript-parse-html-scripts",
                ),
            ]
            try:
                for pattern, parser, parser_name in patterns:
                    if any(fnmatch(value, x) for x in pattern):
                        observer.on_next(
                            Request(
                                real_path=value,
                                parser=parser,
                                parser_name=parser_name,
                            ),
                        )
                        break
            except Exception as ex:  # noqa: BLE001
                observer.on_error(ex)

        return source.subscribe(
            on_next,
            observer.on_error,
            observer.on_completed,
            scheduler=scheduler,
        )

    return reactivex.create(subscribe)



from pydantic import BaseModel, ConfigDict


class Release(BaseModel):
    id_: str
    version_id: str
    name: str | None = None
    pretty_name: str | None = None
    version: str | None = None
    id_like: list[str] | None = None
    version_code_name: str | None = None
    build_id: str | None = None
    image_id: str | None = None
    image_version: str | None = None
    variant: str | None = None
    variant_id: str | None = None
    home_url: str | None = None
    support_url: str | None = None
    bug_report_url: str | None = None
    privacy_policy_url: str | None = None
    cpe_name: str | None = None
    support_end: str | None = None

    def __str__(self) -> str:
        if self.pretty_name:
            return self.pretty_name
        if self.name:
            return self.name
        if self.version:
            return f"{self.id_} {self.version}"
        if self.version_id != "":
            return f"{self.id_} {self.version_id}"
        return f"{self.id_} {self.build_id or ''}"


class Environment(BaseModel):
    linux_release: Release | None
    model_config = ConfigDict(frozen=True)



from typing import NamedTuple

from bs4 import Tag

from labels.model.ecosystem_data.java import JavaPomProject


class ParsedPomProject(NamedTuple):
    java_pom_project: JavaPomProject


class PomContext(NamedTuple):
    project: Tag
    dependencies: Tag
    parent_info: dict[str, str] | None
    parent_version_properties: dict[str, str] | None
    manage_deps: dict[str, str] | None



import re
from typing import NamedTuple
from urllib.parse import urljoin

import requests
from bs4 import BeautifulSoup, NavigableString, Tag

from labels.config.cache import dual_cache
from labels.model.ecosystem_data.java import JavaPomParent, JavaPomProperties
from labels.parsers.cataloger.java.utils.model import PomContext
from labels.parsers.cataloger.java.utils.parsing_utils import get_next_text


class ParsedPomResult(NamedTuple):
    version: str | None
    parent_info: JavaPomParent | None


def _cached_request(url: str) -> requests.Response:
    return requests.get(url, timeout=30)


def _format_maven_pom_ulr(group_id: str, artifact_id: str, version: str) -> str:
    maven_base_url = "https://repo1.maven.org/maven2"
    artifact_pom = f"{artifact_id}-{version}.pom"
    path_components = [*group_id.split("."), artifact_id, version, artifact_pom]

    url = maven_base_url
    for component in path_components:
        url = urljoin(url + "/", component)

    return url


def resolve_version(
    pom_context: PomContext, pom_properties: JavaPomProperties, full_name: str
) -> str | None:
    version = pom_properties.version

    if version and version.startswith("${"):
        version = _resolve_property_version(
            version, pom_context.project, pom_context.parent_version_properties
        )

    if (
        not version
        and pom_context.parent_info
        and pom_properties.group_id
        and pom_properties.artifact_id
    ):
        version = _resolve_version_from_parent_pom_chain(
            group_id=pom_properties.group_id,
            artifact_id=pom_properties.artifact_id,
            parent_info=pom_context.parent_info,
        )

    if not version and pom_context.manage_deps and pom_context.parent_version_properties:
        version = _resolve_managed_version(
            pom_context.manage_deps,
            pom_context.parent_version_properties,
            full_name,
        )

    return version


def _extract_bracketed_text(item: str) -> str:
    match = re.search(r"\$\{([^}]+)\}", item)
    if match:
        return match.group(1)

    return ""


def _resolve_property_version(
    version: str,
    project: Tag,
    parent_version_properties: dict[str, str] | None,
) -> str | None:
    property_name = _extract_bracketed_text(version)

    property_node = project.find_next(property_name)
    if property_node:
        version_text = property_node.get_text()
        if version_text and not version_text.startswith("${"):
            return version_text

    if parent_version_properties:
        return parent_version_properties.get(property_name)

    return None


def _resolve_managed_version(
    manage_deps: dict[str, str], parent_version_properties: dict[str, str], full_name: str
) -> str | None:
    managed_version = manage_deps.get(full_name)
    if managed_version:
        if not managed_version.startswith("${"):
            return managed_version

        return parent_version_properties.get(_extract_bracketed_text(managed_version))

    return None


@dual_cache
def _resolve_version_from_parent_pom_chain(
    group_id: str,
    artifact_id: str,
    parent_info: dict[str, str],
) -> str | None:
    parent_group_id = parent_info["group"]
    parent_artifact_id = parent_info["artifact"]
    parent_version = parent_info["version"]

    for _ in range(3):
        parent_pom = _get_pom_from_maven_repo(
            group_id=parent_group_id,
            artifact_id=parent_artifact_id,
            version=parent_version,
        )
        if not parent_pom:
            break

        parsed_pom_result = _process_pom(parent_pom, group_id, artifact_id)
        if not parsed_pom_result:
            break

        version = parsed_pom_result.version
        if version:
            return version

        parsed_parent_info = parsed_pom_result.parent_info
        if not parsed_parent_info:  # pragma: no cover
            break

        parent_group_id = parsed_parent_info.group_id
        parent_artifact_id = parsed_parent_info.artifact_id
        parent_version = parsed_parent_info.version

    return None


def _process_pom(parent_pom: Tag, group_id: str, artifact_id: str) -> ParsedPomResult | None:
    project = parent_pom.project
    if not project:
        return None

    dependency_management = project.find_next("dependencymanagement")
    if dependency_management and isinstance(dependency_management, Tag):
        version = _get_dependency_version(dependency_management, group_id, artifact_id)
        if version:
            return ParsedPomResult(version=version, parent_info=None)

    parent = project.find_next("parent")
    if parent:
        parent_info = _get_parent_information(parent)
        if parent_info:
            return ParsedPomResult(version=None, parent_info=parent_info)

    return None


def _get_parent_information(
    parent: Tag | NavigableString,
) -> JavaPomParent | None:
    parent_group_id = get_next_text(parent, "groupid")
    parent_artifact_id = get_next_text(parent, "artifactid")
    parent_version = get_next_text(parent, "version")

    if not parent_group_id or not parent_artifact_id or not parent_version:
        return None

    return JavaPomParent(
        group_id=parent_group_id,
        artifact_id=parent_artifact_id,
        version=parent_version,
    )


def _get_dependency_version(
    dependency_management: Tag, group_id: str, artifact_id: str
) -> str | None:
    dependencies: list[Tag] = dependency_management.find_all("dependency")
    for dependency in dependencies:
        if _matches_dependency(dependency, group_id, artifact_id):
            version_node = dependency.find_next("version")
            if version_node:
                return str(version_node.get_text())

    return None


def _matches_dependency(dependency: Tag, group_id: str, artifact_id: str) -> bool:
    dependency_group_id_node = dependency.find_next("groupid")
    dependency_artifact_id_node = dependency.find_next("artifactid")

    return (
        dependency_group_id_node is not None
        and dependency_artifact_id_node is not None
        and dependency_group_id_node.get_text() == group_id
        and dependency_artifact_id_node.get_text() == artifact_id
    )


def _get_pom_from_maven_repo(
    *, group_id: str, artifact_id: str, version: str
) -> BeautifulSoup | None:
    request_url = _format_maven_pom_ulr(group_id, artifact_id, version)
    request = _cached_request(request_url)
    if request.status_code != 200:
        return None

    pom_text = request.text
    return BeautifulSoup(pom_text, features="html.parser")



from labels.model.file import LocationReadCloser
from labels.model.indexables import IndexedDict, ParsedValue
from labels.model.package import Package
from labels.model.package_manager import PackageManager
from labels.model.relationship import Relationship
from labels.model.release import Environment
from labels.model.resolver import Resolver
from labels.parsers.cataloger.javascript.package_builder import new_simple_npm_package
from labels.parsers.cataloger.utils import get_enriched_location
from labels.parsers.collection.json import parse_json_with_tree_sitter


def parse_package_json(
    _resolver: Resolver | None,
    _environment: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    file_content = parse_json_with_tree_sitter(reader.read_closer.read())
    if not isinstance(file_content, IndexedDict):
        return [], []

    packages = _collect_packages(file_content, reader)

    root_package = _create_root_package(file_content, reader)
    if root_package:
        packages.insert(0, root_package)

    return packages, []


def _create_root_package(
    file_content: IndexedDict[str, ParsedValue], reader: LocationReadCloser
) -> Package | None:
    root_name = file_content.get("name")
    root_version = file_content.get("version")

    if not isinstance(root_name, str) or not isinstance(root_version, str):
        return None

    root_location = get_enriched_location(
        reader.location,
        line=1,
        is_transitive=False,
        is_dev=False,
        package_manager=PackageManager.NPM,
    )

    return new_simple_npm_package(root_location, root_name, root_version)


def _collect_packages(
    file_content: IndexedDict[str, ParsedValue], reader: LocationReadCloser
) -> list[Package]:
    return [
        *_create_packages(file_content, reader, is_dev=False),
        *_create_packages(file_content, reader, is_dev=True),
    ]


def _create_packages(
    content: IndexedDict[str, ParsedValue], reader: LocationReadCloser, *, is_dev: bool
) -> list[Package]:
    packages: list[Package] = []
    dependencies_key = "devDependencies" if is_dev else "dependencies"

    dependencies = content.get(dependencies_key)
    if not isinstance(dependencies, IndexedDict):
        return packages

    for package_name, specifier in dependencies.items():
        if not package_name:
            continue

        new_location = get_enriched_location(
            reader.location,
            line=dependencies.get_key_position(package_name).start.line,
            is_dev=is_dev,
            is_transitive=False,
            package_manager=PackageManager.NPM,
        )

        package = new_simple_npm_package(new_location, package_name, str(specifier))
        if package:
            packages.append(package)

    return packages



from labels.model.file import LocationReadCloser
from labels.model.indexables import IndexedDict, IndexedList, ParsedValue
from labels.model.package import Package
from labels.model.package_manager import PackageManager
from labels.model.relationship import Relationship
from labels.model.release import Environment
from labels.model.resolver import Resolver
from labels.parsers.cataloger.swift.package_builder import new_swift_package_manager_package
from labels.parsers.cataloger.utils import get_enriched_location
from labels.parsers.collection.json import parse_json_with_tree_sitter


def parse_package_resolved(
    _resolver: Resolver | None,
    _environment: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    package_resolved = parse_json_with_tree_sitter(reader.read_closer.read())
    if not isinstance(package_resolved, IndexedDict):
        return [], []

    package_resolved_pins: ParsedValue = package_resolved.get("pins")
    if not isinstance(package_resolved_pins, IndexedList):
        return [], []

    packages = _collect_packages(package_resolved_pins, reader)

    return packages, []


def _collect_packages(
    package_resolved_pins: IndexedList[ParsedValue], reader: LocationReadCloser
) -> list[Package]:
    packages: list[Package] = []
    for pin in package_resolved_pins:
        if not isinstance(pin, IndexedDict):
            continue
        info = _get_name_and_version(pin)
        if not info:
            continue

        new_location = get_enriched_location(
            reader.location,
            line=pin.get_key_position("identity").start.line,
            package_manager=PackageManager.SWIFT_PM,
        )
        if pkg := new_swift_package_manager_package(
            source_url=info[0], version=info[1], location=new_location
        ):
            packages.append(pkg)

    return packages


def _get_name_and_version(
    pin: IndexedDict[str, ParsedValue],
) -> tuple[str, str] | None:
    state: ParsedValue = pin.get("state")

    if not isinstance(state, IndexedDict):
        return None

    name = pin.get("location")
    version = state.get("version")

    if not isinstance(version, str) or not isinstance(name, str):
        return None

    return name, version



from more_itertools import flatten

from labels.model.file import Location, LocationReadCloser
from labels.model.indexables import IndexedDict, ParsedValue
from labels.model.package import Ecosystem, Package
from labels.model.package_manager import PackageManager
from labels.model.relationship import Relationship
from labels.model.release import Environment
from labels.model.resolver import Resolver
from labels.parsers.cataloger.dotnet.package_builder import new_dotnet_package
from labels.parsers.cataloger.dotnet.utils import build_relationships_from_json_dependencies
from labels.parsers.cataloger.utils import get_enriched_location
from labels.parsers.collection.json import parse_json_with_tree_sitter
from labels.utils.strings import normalize_name


def parse_dotnet_deps_json(
    _resolver: Resolver | None,
    _environment: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    file_content = parse_json_with_tree_sitter(reader.read_closer.read())
    if not isinstance(file_content, IndexedDict):
        return [], []

    targets = _collect_targets(file_content)
    if not targets:
        return [], []

    packages = _collect_packages(targets, reader.location)
    relationships = _collect_relationships(targets, packages)

    return packages, relationships


def _collect_targets(
    file_content: IndexedDict[str, ParsedValue],
) -> list[tuple[str, ParsedValue]] | None:
    targets: ParsedValue = file_content.get("targets")
    if not isinstance(targets, IndexedDict):
        return None

    return list(
        flatten(target.items() for target in targets.values() if isinstance(target, IndexedDict))
    )


def _collect_packages(targets: list[tuple[str, ParsedValue]], location: Location) -> list[Package]:
    packages: list[Package] = []

    for package_key, package_value in targets:
        if not isinstance(package_value, IndexedDict):
            continue

        package_info = _get_validated_package_info(package_key)
        if not package_info:
            continue

        name, version = package_info
        new_location = get_enriched_location(
            location,
            line=package_value.position.start.line,
            package_manager=PackageManager.NUGET,
        )

        new_package = new_dotnet_package(name, version, new_location)
        if new_package:
            packages.append(new_package)

    return packages


def _collect_relationships(
    targets: list[tuple[str, ParsedValue]],
    packages: list[Package],
) -> list[Relationship]:
    relationships: list[Relationship] = []

    packages_by_key = {(pkg.name, pkg.version): pkg for pkg in packages}

    for package_key, package_value in targets:
        relationships.extend(
            _build_relationships_for_package(package_key, package_value, packages_by_key)
        )

    return relationships


def _build_relationships_for_package(
    package_key: str,
    package_value: ParsedValue,
    packages_by_key: dict[tuple[str, str], Package],
) -> list[Relationship]:
    relationships: list[Relationship] = []

    if not isinstance(package_value, IndexedDict):
        return relationships

    package_info = _get_validated_package_info(package_key)
    if not package_info:
        return relationships

    raw_name, version = package_info
    normalized_name = normalize_name(raw_name, Ecosystem.NUGET)

    current_package = packages_by_key.get((normalized_name, version))
    if not isinstance(current_package, Package):
        return relationships

    dependencies = package_value.get("dependencies")
    if not isinstance(dependencies, IndexedDict):
        return relationships

    return build_relationships_from_json_dependencies(
        dependencies, current_package, packages_by_key
    )


def _get_validated_package_info(package_key: str) -> tuple[str, str] | None:
    if not _is_valid_package_key(package_key):
        return None

    return _split_package_key_into_name_and_version(package_key)


def _is_valid_package_key(package_key: str) -> bool:
    return isinstance(package_key, str) and "/" in package_key


def _split_package_key_into_name_and_version(package_key: str) -> tuple[str, str]:
    name, version = package_key.split("/", 1)
    return name, version



from collections.abc import Callable

import reactivex
from reactivex import Observable
from reactivex.abc import ObserverBase, SchedulerBase

from labels.model.parser import Request, Task
from labels.model.resolver import Resolver


def gen_location_tasks(
    resolver: Resolver,
) -> Callable[[Observable[Request]], Observable[Task]]:
    def _handle(source: Observable[Request]) -> Observable[Task]:
        def subscribe(
            observer: ObserverBase[Task],
            scheduler: SchedulerBase | None = None,
        ) -> reactivex.abc.DisposableBase:
            def on_next(value: Request) -> None:
                try:
                    locations = resolver.files_by_path(value.real_path)
                    for location in locations:
                        observer.on_next(
                            Task(
                                location=location,
                                parser=value.parser,
                                parser_name=value.parser_name,
                            ),
                        )
                except Exception as ex:  # noqa: BLE001
                    observer.on_error(ex)

            return source.subscribe(
                on_next,
                observer.on_error,
                observer.on_completed,
                scheduler=scheduler,
            )

        return reactivex.create(subscribe)

    return _handle



import base64
import logging
from typing import Any, cast

import aioboto3
import botocore

from labels.model.sources import AwsCredentials, AwsRole
from labels.utils.exceptions import AwsCredentialsNotFoundError

LOGGER = logging.getLogger(__name__)


async def get_credentials(credentials: AwsRole) -> AwsCredentials | None:
    try:
        async with aioboto3.Session().client(service_name="sts") as sts_client:
            response = await sts_client.assume_role(
                ExternalId=credentials.external_id,
                RoleArn=credentials.role,
                RoleSessionName="FluidAttacksRoleVerification",
            )
            aws_credentials = AwsCredentials(
                access_key_id=response["Credentials"]["AccessKeyId"],
                secret_access_key=response["Credentials"]["SecretAccessKey"],
                session_token=response["Credentials"]["SessionToken"],
            )
    except botocore.exceptions.ClientError:  # type: ignore[misc]
        LOGGER.exception("We found problems on your AWS credentials")
        raise
    else:
        return aws_credentials


async def obtain_auth_data(  # type: ignore[explicit-any]
    credentials: AwsCredentials,
    region: str = "us-east-1",
) -> dict[str, Any]:
    try:
        session = aioboto3.Session(
            aws_access_key_id=credentials.access_key_id,
            aws_secret_access_key=credentials.secret_access_key,
            aws_session_token=credentials.session_token,
        )
        async with session.client("ecr", region) as client:
            authorization_data = await client.get_authorization_token()
            return cast("dict[str, Any]", authorization_data)  # type: ignore[explicit-any, misc]
    except botocore.exceptions.ClientError:  # type: ignore[misc]
        LOGGER.exception("We found problems on your AWS credentials")
        raise


async def get_token(credentials: AwsCredentials, region: str = "us-east-1") -> str:
    authorization_data = await obtain_auth_data(credentials=credentials, region=region)  # type: ignore[misc]
    auth_resp = authorization_data["authorizationData"][0]  # type: ignore[misc]
    token = auth_resp["authorizationToken"]  # type: ignore[misc]
    _, password = base64.b64decode(token).decode("utf-8").split(":")  # type: ignore[misc]
    return password


async def ecr_connection(role: AwsRole, region: str = "us-east-1") -> str:
    credentials = await get_credentials(role)

    if not credentials:
        raise AwsCredentialsNotFoundError

    token = await get_token(credentials, region=region)
    return f"AWS:{token}"



import logging
import multiprocessing
import traceback

import reactivex
from reactivex import Observable
from reactivex import operators as ops
from reactivex.scheduler import ThreadPoolScheduler

from labels.model.package import Ecosystem, Package
from labels.model.relationship import Relationship
from labels.model.release import Environment
from labels.model.resolver import Resolver
from labels.model.syft_sbom import SyftSBOM
from labels.parsers.builder.builder import build_package
from labels.parsers.builder.relationships_builder import get_relationships
from labels.parsers.operations.cataloger import execute_parsers
from labels.parsers.operations.generator import gen_location_tasks
from labels.parsers.operations.handler import handle_parser
from labels.parsers.operations.utils import handle_relationships, identify_release

LOGGER = logging.getLogger(__name__)


def log_and_continue(
    e: Exception, file_item: str
) -> Observable[tuple[list[Package], list[Relationship]]]:
    LOGGER.error(
        "Error found while resolving packages of %s: %s: %s",
        file_item,
        str(e),
        traceback.format_exc(),
    )
    return reactivex.empty()  # type: ignore[misc]


def process_file_item(
    file_item: str,
    resolver: Resolver,
    pool_scheduler: ThreadPoolScheduler,
) -> Observable[tuple[list[Package], list[Relationship]]]:
    return reactivex.just(file_item).pipe(
        handle_parser(scheduler=pool_scheduler),
        gen_location_tasks(resolver),
        execute_parsers(resolver, Environment(linux_release=identify_release(resolver))),
        ops.catch(lambda e, _: log_and_continue(e, file_item)),
    )


# Syft package types: https://github.com/anchore/syft/blob/main/syft/pkg/type.go
# Syft catalogers: https://github.com/anchore/syft/tree/main/syft/pkg/cataloger
# To add a new type: add mapping here and a builder in BUILDERS (builder.py)
_SYFT_TYPE_TO_ECOSYSTEM: dict[str, Ecosystem] = {
    "alpm": Ecosystem.ALPM,
    "apk": Ecosystem.ALPINE,
    "pod": Ecosystem.SWIFTURL,
    "dart-pub": Ecosystem.PUB,
    "deb": Ecosystem.DEBIAN,
    "dotnet": Ecosystem.NUGET,
    "gem": Ecosystem.RUBYGEMS,
    "go-module": Ecosystem.GO,
    "java-archive": Ecosystem.MAVEN,
    "npm": Ecosystem.NPM,
    "php-composer": Ecosystem.PACKAGIST,
    "php-pear": Ecosystem.PACKAGIST,
    "python": Ecosystem.PYPI,
    "rpm": Ecosystem.RPM,
    "swift": Ecosystem.SWIFTURL,
}


def _get_ecosystem(syft_type: str) -> Ecosystem | None:
    return _SYFT_TYPE_TO_ECOSYSTEM.get(syft_type)


def package_operations_from_sbom(
    docker_sbom: SyftSBOM,
) -> tuple[list[Package], list[Relationship]]:
    packages: list[Package] = []

    for artifact_pkg in docker_sbom.artifacts:
        ecosystem = _get_ecosystem(artifact_pkg.type)
        if ecosystem is None:
            continue

        package = build_package(
            artifact=artifact_pkg,
            ecosystem=ecosystem,
        )

        if package:
            packages.append(package)

    relations: list[Relationship] = get_relationships(docker_sbom, packages)

    return packages, relations


def package_operations_factory(
    resolver: Resolver,
) -> tuple[list[Package], list[Relationship]]:
    observer: Observable[str] = reactivex.from_iterable(resolver.walk_file())
    result_packages: list[Package] = []
    result_relations: list[Relationship] = []
    completed_event = multiprocessing.Event()
    errors = []

    def on_completed() -> None:
        completed_event.set()

    def on_error(error: Exception) -> None:
        errors.append(error)
        on_completed()

    def on_next(value: tuple[list[Package], list[Relationship]]) -> None:
        packages, relations = value
        result_packages.extend(packages)
        result_relations.extend(relations)

    optimal_thread_count = multiprocessing.cpu_count()
    pool_scheduler = ThreadPoolScheduler(optimal_thread_count)

    final_obs: Observable[tuple[list[Package], list[Relationship]]] = observer.pipe(
        ops.map(
            lambda file_item: process_file_item(file_item, resolver, pool_scheduler),  # type: ignore[arg-type]
        ),
        ops.merge(max_concurrent=optimal_thread_count),  # type: ignore[misc]
    )
    final_obs.subscribe(on_next=on_next, on_error=on_error, on_completed=on_completed)

    completed_event.wait()
    result_relations.extend(handle_relationships(result_packages))

    return result_packages, result_relations



import os
from pathlib import Path

CI_COMMIT_REF_NAME = os.environ.get("CI_COMMIT_REF_NAME", "local")
CI_COMMIT_SHA = os.environ.get("CI_COMMIT_SHA", "0000000000000000000000000000000000000000")
CI_COMMIT_SHORT_SHA = CI_COMMIT_SHA[0:8]
BASE_DIR = Path(__file__).resolve().parent



from labels.model.ecosystem_data.base import EcosystemDataModel


class AlpmDBEntry(EcosystemDataModel):
    base_package: str = ""
    package: str = ""
    version: str = ""
    architecture: str = ""
    packager: str = ""



from typing import NamedTuple

from packageurl import PackageURL
from pydantic import ValidationError

from labels.model.file import Location
from labels.model.package import Ecosystem, Language, Package
from labels.parsers.cataloger.utils import log_malformed_package_warning


class DartPubspecLickEntry(NamedTuple):
    hosted_url: str
    vcs_url: str


def new_pubspec_package(
    name: str | None,
    version: str | None,
    location: Location,
    metadata: DartPubspecLickEntry | None = None,
) -> Package | None:
    if not name or not version:
        return None

    p_url = _package_url_for_pubspec(name, version, metadata)

    try:
        return Package(
            name=name,
            version=version,
            locations=[location],
            language=Language.DART,
            ecosystem=Ecosystem.PUB,
            p_url=p_url,
        )
    except ValidationError as ex:
        log_malformed_package_warning(location, ex)
        return None


def _package_url_for_pubspec(
    name: str, version: str, ecosystem_data: DartPubspecLickEntry | None = None
) -> str:
    qualifiers: dict[str, str] = {}
    if ecosystem_data:
        if ecosystem_data.hosted_url:
            qualifiers["hosted_url"] = ecosystem_data.hosted_url
        elif ecosystem_data.vcs_url:
            qualifiers["vcs_url"] = ecosystem_data.vcs_url

    return PackageURL(type="pub", name=name, version=version, qualifiers=qualifiers).to_string()



import hashlib
import json
from enum import StrEnum
from typing import Any, cast

from pydantic import BaseModel

from labels.model.advisories import Advisory
from labels.model.constraints import NonEmptyStr, TrimmedNonEmptyStr
from labels.model.ecosystem_data.aliases import AcceptedEcosystemData
from labels.model.file import Location
from labels.model.metadata import HealthMetadata


class Ecosystem(StrEnum):
    ALPM = "alpm"
    ALPINE = "alpine"
    DEBIAN = "debian"
    GO = "go"
    MAVEN = "maven"
    NPM = "npm"
    NUGET = "nuget"
    PACKAGIST = "packagist"
    PUB = "pub"
    PYPI = "pypi"
    RPM = "rpm"
    RUBYGEMS = "rubygems"
    SWIFTURL = "swifturl"


class Language(StrEnum):
    UNKNOWN_LANGUAGE = "unknown_language"
    DART = "dart"
    DOTNET = "dotnet"
    GO = "go"
    JAVA = "java"
    JAVASCRIPT = "javascript"
    PHP = "php"
    PYTHON = "python"
    RUBY = "ruby"
    SWIFT = "swift"


class Package(BaseModel):
    name: NonEmptyStr
    version: TrimmedNonEmptyStr
    language: Language
    locations: list[Location]
    ecosystem: Ecosystem
    advisories: list[Advisory] | None = None
    found_by: NonEmptyStr | None = None
    health_metadata: HealthMetadata | None = None
    ecosystem_data: AcceptedEcosystemData | None = None
    p_url: NonEmptyStr
    safe_versions: list[str] | None = None
    syft_id: NonEmptyStr | None = None

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Package):
            return False

        self_dump = cast("dict[str, object]", self.model_dump())
        other_dump = cast("dict[str, object]", other.model_dump())

        return self_dump == other_dump

    def __hash__(self) -> int:
        return hash(self.id_)

    @property
    def id_(self) -> str:
        return self.id_by_hash()

    def id_by_hash(self) -> str:
        try:
            obj_data = {
                "name": self.name,
                "version": self.version,
                "language": self.language.value,
                "ecosystem": self.ecosystem.value,
                "p_url": self.p_url,
            }
            obj_str = json.dumps(obj_data, sort_keys=True)
            return hashlib.sha256(obj_str.encode()).hexdigest()
        except Exception as exc:
            error_message = "Package data invalid for ID generation: " + json.dumps(
                obj_data, sort_keys=True
            )
            raise ValueError(error_message) from exc

    def serialize(self) -> dict[str, Any]:  # type: ignore[explicit-any]
        return {
            "id": self.id_,
            "name": self.name,
            "version": self.version,
            "locations": [loc.serialize() for loc in self.locations],  # type: ignore[misc]
            "language": self.language.value,
            "ecosystem": self.ecosystem.value,
            "package_url": self.p_url,
            "found_by": self.found_by,
            "health_metadata": self.health_metadata.model_dump() if self.health_metadata else None,  # type: ignore[misc]
            "safe_versions": self.safe_versions,
            "advisories": [adv.serialize() for adv in self.advisories or []],  # type: ignore[misc]
        }



from labels.model.file import LocationReadCloser
from labels.model.indexables import IndexedDict
from labels.model.package import Package
from labels.model.relationship import Relationship
from labels.model.release import Environment
from labels.model.resolver import Resolver
from labels.parsers.cataloger.javascript.parse_package_lock.parse_package_lock_v1 import (
    parse_package_lock_v1,
)
from labels.parsers.cataloger.javascript.parse_package_lock.parse_package_lock_v2 import (
    parse_package_lock_v2,
)
from labels.parsers.collection.json import parse_json_with_tree_sitter


def parse_package_lock(
    _resolver: Resolver | None,
    _environment: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    file_content = parse_json_with_tree_sitter(reader.read_closer.read())
    if not isinstance(file_content, IndexedDict):
        return [], []

    match file_content.get("lockfileVersion", 1):
        case 1:
            return parse_package_lock_v1(reader.location, file_content)
        case 2 | 3:
            return parse_package_lock_v2(reader.location, file_content)

    return [], []



from bs4 import NavigableString, Tag


def get_next_text(parent: Tag | NavigableString, name: str) -> str | None:
    element = parent.find_next(name)
    if element:
        return element.get_text()

    return None


def get_direct_child_text(parent: Tag, name: str) -> str | None:
    element = parent.find(name)
    if element:
        return element.get_text()

    return None


def safe_get_text(value: Tag | None) -> str:
    if not value:
        return ""
    return value.get_text()



from concurrent.futures import ThreadPoolExecutor
from multiprocessing import cpu_count

from spdx_tools.spdx.model.document import Document
from spdx_tools.spdx.model.package import Package as SPDX_Package
from spdx_tools.spdx.model.package import PackagePurpose
from spdx_tools.spdx.model.relationship import Relationship as SPDXRelationship
from spdx_tools.spdx.model.relationship import RelationshipType as SPDXRelationshipType

from labels.model.package import Package
from labels.model.relationship import Relationship
from labels.output.spdx.complete_file import (
    NOASSERTION,
    add_external_refs,
    add_integrity,
    get_spdx_id,
)


def package_to_spdx_pkg(package: Package) -> SPDX_Package:
    spdx_id = get_spdx_id(package)
    external_refs = add_external_refs(package)
    checksums = add_integrity(package.health_metadata)

    return SPDX_Package(
        spdx_id=spdx_id,
        name=package.name,
        download_location=NOASSERTION,
        version=package.version,
        primary_package_purpose=PackagePurpose.LIBRARY,
        external_references=external_refs,
        checksums=checksums,
    )


def create_package_cache(packages: list[Package]) -> dict[str, SPDX_Package]:
    with ThreadPoolExecutor(max_workers=cpu_count()) as executor:
        spdx_packages = list(executor.map(package_to_spdx_pkg, packages))
    return {pkg.id_: spdx_pkg for pkg, spdx_pkg in zip(packages, spdx_packages, strict=True)}


def create_document_relationships(
    document: Document,
    spdx_packages: list[SPDX_Package],
) -> list[SPDXRelationship]:
    doc_spdx_id = document.creation_info.spdx_id
    return [
        SPDXRelationship(doc_spdx_id, SPDXRelationshipType.DESCRIBES, pkg.spdx_id)
        for pkg in spdx_packages
    ]


def process_relationship(
    relationship: Relationship,
    spdx_id_cache: dict[str, str],
) -> SPDXRelationship | None:
    if relationship.type.value != "dependency-of":
        return None

    # Extract base IDs (remove @location_id suffix if present)
    to_id = relationship.to_.split("@")[0] if "@" in relationship.to_ else relationship.to_
    from_id = relationship.from_.split("@")[0] if "@" in relationship.from_ else relationship.from_

    to_pkg_id = spdx_id_cache.get(to_id)
    from_pkg_id = spdx_id_cache.get(from_id)

    if to_pkg_id and from_pkg_id:
        return SPDXRelationship(to_pkg_id, SPDXRelationshipType.DEPENDENCY_OF, from_pkg_id)

    return None


def process_relationships(
    relationships: list[Relationship],
    spdx_id_cache: dict[str, str],
) -> list[SPDXRelationship]:
    with ThreadPoolExecutor(max_workers=cpu_count()) as executor:
        processed_relationships = list(
            executor.map(lambda r: process_relationship(r, spdx_id_cache), relationships),
        )

    return [relationship for relationship in processed_relationships if relationship is not None]


def add_packages_and_relationships(
    document: Document,
    packages: list[Package],
    relationships: list[Relationship],
) -> None:
    package_cache = create_package_cache(packages)
    spdx_packages = list(package_cache.values())
    spdx_id_cache = {pkg_id: spdx_pkg.spdx_id for pkg_id, spdx_pkg in package_cache.items()}

    document.packages = spdx_packages
    document.relationships.extend(
        [
            *create_document_relationships(document, document.packages),
            *process_relationships(relationships, spdx_id_cache),
        ],
    )



import re
from collections.abc import Callable

from labels.model.package import Ecosystem

UNSTABLE_SEMVER_KEYWORDS: frozenset[str] = frozenset(
    {
        "alpha",
        "beta",
        "canary",
        "cr",
        "dev",
        "milestone",
        "next",
        "nightly",
        "pre",
        "preview",
        "rc",
        "snapshot",
    }
)

STABLE_QUALIFIERS: frozenset[str] = frozenset(
    {
        "final",
        "ga",
        "ga.final",
        "ga.release",
        "lts",
        "release",
        "release.final",
        "stable",
    }
)

_OS_ECOSYSTEMS: frozenset[Ecosystem] = frozenset(
    {
        Ecosystem.ALPM,
        Ecosystem.ALPINE,
        Ecosystem.DEBIAN,
        Ecosystem.RPM,
    }
)

_SEMVER_PRE_RELEASE_RE = re.compile(r"^\d+(?:\.\d+)*-(.+?)(?:\+.*)?\Z")
_SEMVER_QUALIFIER_RE = re.compile(r"^\d+(?:\.\d+)*\.([a-zA-Z][a-zA-Z0-9.]*)\Z")

_PEP440_PRE_RE = re.compile(
    r"^\d+(?:\.\d+)*"
    r"(?:\.?(?:a|alpha|b|beta|c|rc)\d*)"
    r"(?:\..*)?(?:\+.*)?\Z",
    re.IGNORECASE,
)
_PEP440_DEV_RE = re.compile(
    r"^\d+(?:\.\d+)*\.?dev\d*(?:\+.*)?\Z",
    re.IGNORECASE,
)

_RUBY_PRE_RE = re.compile(
    r"^\d+(?:\.\d+)*\."
    r"(?:alpha|beta|dev|pre|rc)\d*(?:\.\d+)*\Z",
    re.IGNORECASE,
)

_M_DIGIT_RE = re.compile(r"^m\d+\Z", re.IGNORECASE)


def _is_unstable_keyword(segment: str) -> bool | None:
    if segment in STABLE_QUALIFIERS:
        return False
    if segment in UNSTABLE_SEMVER_KEYWORDS:
        return True
    if _M_DIGIT_RE.match(segment):
        return True
    base = segment.rstrip("0123456789")
    if base and base != segment and base in UNSTABLE_SEMVER_KEYWORDS:
        return True
    return None


def _check_pre_release(pre_release: str) -> bool:
    normalized = pre_release.lower().replace("-", ".").replace("_", ".")
    return any(_is_unstable_keyword(seg) is True for seg in normalized.split("."))


def _is_unstable_semver(version: str) -> bool:
    match = _SEMVER_PRE_RELEASE_RE.match(version)
    if match:
        return _check_pre_release(match.group(1))

    match = _SEMVER_QUALIFIER_RE.match(version)
    if match:
        qualifier: str = match.group(1).lower()
        result = _is_unstable_keyword(qualifier)
        if result is not None:
            return result
        first_segment: str = qualifier.split(".")[0]
        return _is_unstable_keyword(first_segment) or False

    return False


def _is_unstable_python(version: str) -> bool:
    if _PEP440_DEV_RE.match(version):
        return True
    return bool(_PEP440_PRE_RE.match(version))


def _is_unstable_ruby(version: str) -> bool:
    return bool(_RUBY_PRE_RE.match(version))


def _is_unstable_generic(version: str) -> bool:
    if _is_unstable_semver(version):
        return True
    if _is_unstable_python(version):
        return True
    return _is_unstable_ruby(version)


_APP_ECOSYSTEM_DETECTOR: dict[Ecosystem, Callable[[str], bool]] = {
    Ecosystem.PACKAGIST: _is_unstable_semver,
    Ecosystem.RUBYGEMS: _is_unstable_ruby,
    Ecosystem.GO: _is_unstable_semver,
    Ecosystem.MAVEN: _is_unstable_semver,
    Ecosystem.NPM: _is_unstable_semver,
    Ecosystem.NUGET: _is_unstable_semver,
    Ecosystem.PYPI: _is_unstable_python,
    Ecosystem.PUB: _is_unstable_semver,
    Ecosystem.SWIFTURL: _is_unstable_semver,
}


def is_unstable_version(version: str, ecosystem: Ecosystem | None = None) -> bool:
    if not version or "${" in version:
        return False

    cleaned = version.strip().removeprefix("v").removeprefix("V")
    if not cleaned:
        return False

    if ecosystem is not None:
        if ecosystem in _OS_ECOSYSTEMS:
            return False
        detector = _APP_ECOSYSTEM_DETECTOR.get(ecosystem)
        if detector is not None:
            return detector(cleaned)
        return False

    return _is_unstable_generic(cleaned)



from packageurl import PackageURL
from pydantic import ValidationError

from labels.model.file import Location
from labels.model.indexables import IndexedDict, ParsedValue
from labels.model.metadata import HealthMetadata
from labels.model.package import Ecosystem, Language, Package
from labels.parsers.cataloger.utils import log_malformed_package_warning


def new_npm_package_from_lock(
    location: Location,
    name: str | None,
    value: IndexedDict[str, ParsedValue],
    lockfile_version: int,
    health_metadata: HealthMetadata | None = None,
) -> Package | None:
    version: str = str(value.get("version", ""))

    if not name or not version:
        return None

    if lockfile_version == 1:
        alias_prefix_package_lock = "npm:"
        if version.startswith(alias_prefix_package_lock):
            name, version = version.removeprefix(alias_prefix_package_lock).rsplit(
                "@",
                1,
            )

    p_url = _get_npm_package_url(name, version)

    try:
        return Package(
            name=name,
            version=version,
            locations=[location],
            language=Language.JAVASCRIPT,
            ecosystem=Ecosystem.NPM,
            p_url=p_url,
            health_metadata=health_metadata,
        )
    except ValidationError as ex:
        log_malformed_package_warning(location, ex)
        return None


def new_simple_npm_package(
    location: Location,
    name: str | None,
    version: str | None,
    health_metadata: HealthMetadata | None = None,
) -> Package | None:
    if not name or not version:
        return None

    p_url = _get_npm_package_url(name, version)

    try:
        return Package(
            name=name,
            version=version,
            locations=[location],
            language=Language.JAVASCRIPT,
            ecosystem=Ecosystem.NPM,
            p_url=p_url,
            health_metadata=health_metadata,
        )
    except ValidationError as ex:
        log_malformed_package_warning(location, ex)
        return None


def _get_npm_package_url(name: str, version: str) -> str:
    namespace = ""
    fields = name.split("/", 2)
    if len(fields) > 1:
        namespace = fields[0]
        name = fields[1]

    if not name:
        return ""

    return PackageURL(type="npm", namespace=namespace, name=name, version=version).to_string()



from labels.model.file import LocationReadCloser
from labels.model.indexables import IndexedDict, ParsedValue
from labels.model.package import Package
from labels.model.package_manager import PackageManager
from labels.model.relationship import Relationship
from labels.model.release import Environment
from labels.model.resolver import Resolver
from labels.parsers.cataloger.python.package_builder import new_python_package
from labels.parsers.cataloger.utils import get_enriched_location
from labels.parsers.collection import toml


def parse_pipfile_deps(
    _resolver: Resolver | None,
    _environment: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    file_content = toml.parse_toml_with_tree_sitter(reader.read_closer.read())

    packages = _collect_packages(reader, file_content)

    return packages, []


def _collect_packages(
    reader: LocationReadCloser, file_content: IndexedDict[str, ParsedValue]
) -> list[Package]:
    packages: list[Package] = []

    toml_packages = file_content.get("packages")
    if not isinstance(toml_packages, IndexedDict):
        return packages
    packages = _get_packages(reader, toml_packages)

    dev_deps = file_content.get("dev-packages")
    if isinstance(dev_deps, IndexedDict):
        dev_pkgs = _get_packages(reader, dev_deps, is_dev=True)
        packages.extend(dev_pkgs)

    return packages


def _get_packages(
    reader: LocationReadCloser,
    toml_packages: IndexedDict[str, ParsedValue],
    *,
    is_dev: bool = False,
) -> list[Package]:
    result = []
    for raw_package, version_data in toml_packages.items():
        version: str = ""
        if isinstance(version_data, str):
            version = version_data.strip("=<>~^ ")
        if isinstance(version_data, IndexedDict):
            version = str(version_data.get("version", "*")).strip("=<>~^ ")

        if "*" in version:
            continue

        new_location = get_enriched_location(
            reader.location,
            line=raw_package.position.start.line
            if isinstance(raw_package, IndexedDict)
            else toml_packages.get_key_position(raw_package).start.line,
            is_dev=is_dev,
            is_transitive=False,
            package_manager=PackageManager.PIPENV,
        )

        package = new_python_package(name=raw_package, version=version, location=new_location)
        if package:
            result.append(package)

    return result



from packageurl import PackageURL
from pydantic import ValidationError

from labels.model.file import Location
from labels.model.package import Ecosystem, Language, Package
from labels.parsers.cataloger.utils import log_malformed_package_warning


def new_gem_package(name: str | None, version: str | None, location: Location) -> Package | None:
    if not name or not version:
        return None

    p_url = PackageURL(type="gem", name=name, version=version).to_string()

    try:
        return Package(
            name=name,
            version=version,
            ecosystem=Ecosystem.RUBYGEMS,
            locations=[location],
            p_url=p_url,
            language=Language.RUBY,
        )
    except ValidationError as ex:
        log_malformed_package_warning(location, ex)
        return None



import logging
import re
from collections.abc import Iterator
from contextlib import suppress
from typing import cast

import tree_sitter_toml
from more_itertools import mark_ends
from tree_sitter import Language as TLanguage
from tree_sitter import Node, Parser

from labels.model.indexables import IndexedDict, IndexedList, ParsedValue
from labels.utils.exceptions import DuplicatedKeyError, InvalidTypeError, UnexpectedNodeError

LOGGER = logging.getLogger(__name__)
EMPTY_DICT: IndexedDict[str, ParsedValue] = IndexedDict()


def _validate_key_pairs(root_node: Node) -> bool:
    all_children_pairs = tuple(
        node.start_point[0] for node in root_node.children if node.type == "pair"
    )
    return len(all_children_pairs) == len(set(all_children_pairs))


def handle_string(node: Node) -> str:
    value = node.text.decode("utf-8") if node.text else ""
    str_quote = "'" if value.startswith("'") else None
    str_quote = str_quote or ('"' if value.startswith('"') else None)
    if str_quote and (value.startswith(str_quote)) and (value.endswith(str_quote)):
        value = value.strip(str_quote)
        # Remove surrounding triple quotes if present
        value = re.sub(r'"""|\'\'\'', "", value)

        value = value.strip()

    return value


def handle_array(node: Node) -> IndexedList[ParsedValue]:
    data: IndexedList[ParsedValue] = IndexedList(node)
    for children_node in node.children:
        if children_node.type in ("[", "]", ",", "comment"):
            continue
        node_results = handle_node(children_node)
        if not isinstance(node_results, tuple):
            data.append((node_results, children_node))
    return data


def handle_bare_key(node: Node) -> str:
    return handle_string(node)


def handle_quoted_key(node: Node) -> str:
    value = handle_string(node)
    if value.startswith(('"', "'")):
        return value[1:-1]
    return value


def handle_dotted_key(node: Node) -> list[str]:
    value = handle_string(node)
    parts = cast("list[str]", re.findall(r'"[^"]*"|[\w-]+', value))
    return [part.strip('" ') for part in parts]


def handle_table_array_element(
    node: Node,
) -> tuple[tuple[ParsedValue, Node], IndexedDict[str, ParsedValue]]:
    data: IndexedDict[str, ParsedValue] = IndexedDict(node)
    bare_key = node.named_children[0]
    bare_key_value = handle_node(bare_key)
    if isinstance(bare_key_value, tuple):
        return (None, bare_key), data

    # the children's are pair nods
    for children_node in node.named_children[1:]:
        if children_node.type in ("[", "]", ", ", "comment"):
            continue
        if children_node.type != "pair":
            LOGGER.error("Unexpected node: %s", children_node)
            continue
        node_results = handle_node(children_node)
        if isinstance(node_results, tuple):
            key, value = node_results
            if isinstance(key[0], str):
                data[(key[0], key[1])] = value
    return (bare_key_value, bare_key), data


def handle_pair(node: Node) -> tuple[tuple[ParsedValue, Node], tuple[ParsedValue, Node]]:
    bare_key_node, value_node = node.named_children[:2]
    bare_key_value = handle_node(bare_key_node)
    value_resolved = handle_node(value_node)
    if isinstance(bare_key_value, tuple) or isinstance(value_resolved, tuple):
        return (None, bare_key_node), (None, value_node)
    return (bare_key_value, bare_key_node), (value_resolved, value_node)


def handle_boolean(node: Node) -> bool:
    value = node.text.decode("utf-8") if node.text else ""
    return value.lower() == "true"


def handle_integer(node: Node) -> int:
    decode_str = node.text.decode("utf-8") if node.text else ""
    with suppress(ValueError):
        return int(decode_str)
    with suppress(ValueError):
        return int(decode_str, 16)
    with suppress(ValueError):
        return int(decode_str, 8)

    error_msg = f"Invalid integer value: {decode_str}"
    raise ValueError(error_msg)


def handle_float(node: Node) -> float:
    decoded_str = node.text.decode("utf-8") if node.text else ""
    with suppress(ValueError):
        return float(decoded_str)

    with suppress(ValueError):
        return float(decoded_str.replace(".", "").lower())

    error_msg = f"Invalid float value: {decoded_str}"
    raise ValueError(error_msg)


def handle_local_date(node: Node) -> str:
    return handle_string(node)


def handle_inline_table(node: Node) -> IndexedDict[str, ParsedValue]:
    data: IndexedDict[str, ParsedValue] = IndexedDict(node)
    for children_node in node.named_children:
        node_results = handle_node(children_node)
        if isinstance(node_results, tuple):
            key, value = node_results
            if isinstance(key[0], str | list):
                nested_dict(
                    data=data,
                    keys=key[0],
                    keys_node=key[1],
                    value=value[0],
                    value_node=value[1],
                )
    return data


def handle_table(node: Node, data_1: IndexedDict[str, ParsedValue]) -> None:
    data_new: IndexedDict[str, ParsedValue] = IndexedDict(node)
    for children_node in node.named_children[1:]:
        if children_node.type == "comment":
            continue
        node_results = handle_node(children_node)
        if isinstance(node_results, tuple):
            key, value = node_results
            if isinstance(key[0], str | list):
                nested_dict(
                    data=data_new,
                    keys=key[0],
                    keys_node=key[1],
                    value=value[0],
                    value_node=value[1],
                )
    bare_key_node = node.named_children[0]
    bare_key_value = handle_node(bare_key_node)
    if isinstance(bare_key_value, list):
        nested_dict(
            data=data_1,
            keys=bare_key_value,
            keys_node=bare_key_node,
            value=data_new,
            value_node=node,
        )
    elif isinstance(bare_key_value, str):
        data_1[(bare_key_value, bare_key_node)] = (
            data_new,
            node,
        )


def handle_node(  # noqa: C901, PLR0912, PLR0915
    node: Node,
) -> ParsedValue | tuple[tuple[ParsedValue, Node], tuple[ParsedValue, Node]]:
    value: ParsedValue | tuple[tuple[ParsedValue, Node], tuple[ParsedValue, Node]] = None
    match node.type:
        case "comment":
            value = None
        case "string":
            value = handle_string(node)
        case "array":
            value = handle_array(node)
        case "pair":
            value = handle_pair(node)
        case "bare_key":
            value = handle_bare_key(node)
        case "quoted_key":
            value = handle_quoted_key(node)
        case "dotted_key":
            value = handle_dotted_key(node)
        case "boolean":
            value = handle_boolean(node)
        case "integer":
            value = handle_integer(node)
        case "float":
            value = handle_float(node)
        case "inline_table":
            value = handle_inline_table(node)
        case "local_date":
            value = handle_local_date(node)
        case _:
            raise UnexpectedNodeError(node.type)

    return value


def _nested_dict_handle_list(data: ParsedValue) -> ParsedValue:
    if not isinstance(data, IndexedDict):
        if not isinstance(data, IndexedList):
            error_msg = "Attempted to extend non-table type"
            raise InvalidTypeError(error_msg)
        data = data[-1]
    return data


def _nested_dict_validate_duplicated_data(
    data: IndexedDict[str, ParsedValue],
    key: str,
    *,
    is_last: bool,
) -> None:
    def _check_for_duplicates(
        data: IndexedDict[str, ParsedValue],
        key: str,
        *,
        is_last: bool,
    ) -> None:
        if is_last and key in data:
            raise DuplicatedKeyError(key)

    try:
        _check_for_duplicates(data, key, is_last=is_last)
    except DuplicatedKeyError:
        LOGGER.exception(
            "Duplicated key detected: %s",
            key,
            extra={
                "extra": {
                    "key": key,
                    "is_last": is_last,
                },
            },
        )


def nested_dict(
    *,
    data: IndexedDict[str, ParsedValue],
    keys: list[str] | str,
    keys_node: Node,
    value: ParsedValue = None,
    value_node: Node | None = None,
) -> None:
    for _, is_last, key in _mark_ends(keys):
        current_data = _nested_dict_handle_list(data)
        if not isinstance(current_data, IndexedDict):
            continue

        _nested_dict_validate_duplicated_data(current_data, key, is_last=is_last)

        if key not in current_data and is_last:
            current_data[(key, keys_node)] = (
                value if value is not None else IndexedDict(),
                value_node or keys_node,
            )
            continue

        if key not in current_data:
            current_data[(key, keys_node)] = (IndexedDict(), value_node or keys_node)

        next_data = _nested_dict_handle_list(current_data[key])
        if not isinstance(next_data, IndexedDict):
            continue

        data = next_data


def _mark_ends(keys: list[str] | str) -> Iterator[tuple[bool, bool, str]]:
    yield from mark_ends(keys if isinstance(keys, list) else [keys])


def _nested_list_handle_list(data: ParsedValue) -> ParsedValue:
    if isinstance(data, IndexedList):
        data = data[-1]
    return data


def _nested_list_handle_fists_keys(
    data: IndexedDict[str, ParsedValue],
    key: str,
    *,
    is_last: bool,
) -> bool:
    return key not in data and not is_last


def _nested_list_handle_last_keys(
    data: IndexedDict[str, ParsedValue],
    key: str,
    *,
    is_last: bool,
) -> bool:
    return key not in data and is_last


def _nested_list_handle_fists_keys_exists(
    data: IndexedDict[str, ParsedValue],
    key: str,
    *,
    is_last: bool,
) -> bool:
    return key in data and not is_last


def _nested_list(  #  noqa: PLR0913
    *,
    data: IndexedDict[str, ParsedValue],
    key: str,
    is_last: bool,
    keys_node: Node,
    build_key: list[str],
    value: ParsedValue = None,
    value_node: Node | None = None,
) -> None:
    if is_last:
        current_data = data[key]
        if isinstance(current_data, IndexedList):
            current_data.append((value, value_node or keys_node))
        else:
            LOGGER.error("Attempted to extend non-table type: %s", ".".join(build_key))


def nested_list(
    *,
    data: IndexedDict[str, ParsedValue],
    keys: list[str] | str,
    keys_node: Node,
    value: ParsedValue = None,
    value_node: Node | None = None,
) -> None:
    build_key = []
    for _, is_last, key in _mark_ends(keys):
        build_key.append(key)
        current_table = _nested_list_handle_list(data)
        if not isinstance(current_table, IndexedDict):
            continue

        if _nested_list_handle_fists_keys(current_table, key, is_last=is_last):
            current_table[(key, keys_node)] = (IndexedDict(), value_node or keys_node)
            next_data = _nested_list_handle_list(current_table[key])
            if not isinstance(next_data, IndexedDict):
                continue
            data = next_data
            continue

        if _nested_list_handle_fists_keys_exists(current_table, key, is_last=is_last):
            next_data = _nested_list_handle_list(current_table[key])
            if not isinstance(next_data, IndexedDict):
                continue
            data = next_data
            continue

        if _nested_list_handle_last_keys(current_table, key, is_last=is_last):
            current_table[(key, keys_node)] = (
                IndexedList(keys_node),
                value_node or keys_node,
            )

        _nested_list(
            data=current_table,
            key=key,
            is_last=is_last,
            keys_node=keys_node,
            build_key=build_key,
            value=value,
            value_node=value_node,
        )


def _process_node(node: Node, data: IndexedDict[str, ParsedValue]) -> None:
    if node.type == "pair":
        node_results = handle_node(node)
        if isinstance(node_results, tuple):  # pragma: no branch
            key, value = node_results
            if isinstance(key[0], str | list):  # pragma: no branch
                nested_dict(
                    data=data,
                    keys=key[0],
                    keys_node=key[1],
                    value=value[0],
                    value_node=value[1],
                )
    elif node.type == "table":
        handle_table(node, data)
    elif node.type == "table_array_element":
        key, table_value = handle_table_array_element(node)
        if isinstance(key[0], str | list):  # pragma: no branch
            nested_list(
                data=data,
                keys=key[0],
                keys_node=key[1],
                value=table_value,
                value_node=node,
            )
    elif node.type == "ERROR":
        raise UnexpectedNodeError(node)


def parse_toml_with_tree_sitter(toml_content: str) -> IndexedDict[str, ParsedValue]:
    parser_language = TLanguage(tree_sitter_toml.language())
    parser = Parser(parser_language)
    result = parser.parse(toml_content.encode("utf-8"))
    data: IndexedDict[str, ParsedValue] = IndexedDict(result.root_node)
    if result.root_node.type != "document":
        return data
    if not _validate_key_pairs(result.root_node):
        LOGGER.error("Invalid document", extra={"extra": {"toml_content": toml_content}})
        return IndexedDict()
    for node in result.root_node.children:
        try:
            _process_node(node=node, data=data)
        except (UnexpectedNodeError, ValueError, InvalidTypeError) as exc:
            LOGGER.exception(
                "Error processing node of type '%s'",
                node.type,
                extra={"extra": {"node_type": node.type, "toml": toml_content}},
                exc_info=exc,
            )
    return data



import logging
from typing import Protocol

from labels.model.core import OutputFormat, SbomConfig
from labels.model.package import Package
from labels.model.relationship import Relationship
from labels.model.resolver import Resolver
from labels.model.syft_sbom import SyftSBOM
from labels.output.cyclonedx.output_handler import format_cyclonedx_sbom
from labels.output.fluid.output_handler import format_fluid_sbom
from labels.output.spdx.output_handler import format_spdx_sbom

LOGGER = logging.getLogger(__name__)


class SbomFormatter(Protocol):
    def __call__(
        self,
        *,
        packages: list[Package],
        relationships: list[Relationship],
        config: SbomConfig,
        resolver: Resolver | SyftSBOM,
    ) -> None: ...


_FORMAT_HANDLERS: dict[OutputFormat, SbomFormatter] = {
    OutputFormat.FLUID_JSON: format_fluid_sbom,
    OutputFormat.CYCLONEDX_JSON: format_cyclonedx_sbom,
    OutputFormat.CYCLONEDX_XML: format_cyclonedx_sbom,
    OutputFormat.SPDX_JSON: format_spdx_sbom,
    OutputFormat.SPDX_XML: format_spdx_sbom,
}


def dispatch_sbom_output(
    *,
    packages: list[Package],
    relationships: list[Relationship],
    config: SbomConfig,
    resolver: Resolver | SyftSBOM,
) -> None:
    handler = _FORMAT_HANDLERS[config.output_format]
    handler(
        packages=packages,
        relationships=relationships,
        config=config,
        resolver=resolver,
    )



import uuid
from datetime import UTC, datetime

from fluidattacks_tracks import Tracks
from fluidattacks_tracks.resources.event import Event

from labels.model.core import SbomConfig

client = Tracks()


def send_event_to_tracks(
    *,
    sbom_config: SbomConfig,
    packages_amount: int,
    relationships_amount: int,
    vulns_summary: dict[str, int],
) -> None:
    client.event.create(
        Event(
            action="CREATE",
            author="unknown",
            date=datetime.now(UTC),
            mechanism="TASK",
            metadata={
                "status": "success",
                "source": sbom_config.source,
                "exclude": sbom_config.exclude,
                "include": sbom_config.include,
                "aws_role": sbom_config.aws_role,
                "docker_user": sbom_config.docker_user,
                "aws_external_id": sbom_config.aws_external_id,
                "source_type": sbom_config.source_type.value,
                "output_format": sbom_config.output_format.value,
                "packages_count": str(packages_amount),
                "relationships_count": str(relationships_amount),
                "vulns_summary": vulns_summary,
            },
            object="LabelsExecution",
            object_id=sbom_config.execution_id or str(uuid.uuid4()).replace("-", ""),
        ),
    )



import re
from collections.abc import Iterable

from gemfileparser import Dependency, GemfileParser

from labels.model.file import Location, LocationReadCloser
from labels.model.package import Package
from labels.model.package_manager import PackageManager
from labels.model.relationship import Relationship
from labels.model.release import Environment
from labels.model.resolver import Resolver
from labels.parsers.cataloger.ruby.package_builder import new_gem_package
from labels.parsers.cataloger.ruby.utils import format_ruby_version_constraints
from labels.parsers.cataloger.utils import get_enriched_location

GEMFILE_DEP: re.Pattern[str] = re.compile(
    r'^\s*(?P<gem>gem [\'"].*?[\'"],?( [\'"][><~=]{0,2}\s?[\d\.]+[\'"],?){0,2})',
)
NOT_PROD_DEP: re.Pattern[str] = re.compile(
    r":group => \[?[:\w\-, ]*(:development|:test)",
)
NOT_PROD_GROUP: re.Pattern[str] = re.compile(r"(\s*)group :(test|development)")
GEM_LOCK_DEP: re.Pattern[str] = re.compile(
    r"^\s{4}(?P<gem>[^\s]*)\s\([^\d]*(?P<version>.*)\)$",
)


def parse_gemfile(
    _resolver: Resolver | None,
    _environment: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    packages = _collect_packages(reader.read_closer.read(), reader)

    return packages, []


def _collect_packages(file_content: str, reader: LocationReadCloser) -> list[Package]:
    packages: list[Package] = []
    in_group_block = False
    end_line = ""

    for line_number, line in enumerate(file_content.splitlines(), 1):
        if in_group_block:
            if line == end_line:
                in_group_block = False
            elif (
                (matched := GEMFILE_DEP.search(line))
                and matched
                and (dependency := _create_dependency(line_number, reader.location, matched, line))
            ):
                packages.append(dependency)
            continue

        if match_group := NOT_PROD_GROUP.search(line):
            in_group_block = True
            end_line = f"{match_group.group(1)}end"
            continue

        if (
            (matched := GEMFILE_DEP.search(line))
            and matched
            and (dependency := _create_dependency(line_number, reader.location, matched, line))
        ):
            packages.append(dependency)

    return packages


def _create_dependency(
    line_number: int,
    location: Location,
    matched: re.Match[str],
    line: str,
) -> Package | None:
    gem_info = GemfileParser.preprocess(matched.group("gem"))[3:]
    product, version = _parse_line(gem_info)

    is_dev = _is_dev_dependency(line)

    new_location = get_enriched_location(
        location,
        line=line_number,
        is_dev=is_dev,
        is_transitive=False,
        package_manager=PackageManager.BUNDLER,
    )

    return new_gem_package(product, version, new_location)


def _parse_line(in_line: str) -> tuple[str, str]:
    line: list[str] = []
    line = in_line.split(",")

    column_list: list[str] = []
    for column in line:
        stripped_column = (
            column.replace("'", "")
            .replace('"', "")
            .replace("%q<", "")
            .replace("(", "")
            .replace(")", "")
            .replace("[", "")
            .replace("]", "")
            .strip()
        )
        column_list.append(stripped_column)

    deps = _match_dep_criteria(column_list)
    product: str = deps.name
    version: str = format_ruby_version_constraints(deps.requirement)
    return product, version


def _match_dep_criteria(
    column_list: Iterable[str],
) -> Dependency:
    dep = Dependency()
    for column in column_list:
        for criteria, criteria_regex in GemfileParser.gemfile_regexes.items():
            match = criteria_regex.match(column)
            if match:
                if criteria == "requirement":
                    dep.requirement.append(match.group(criteria))
                else:
                    setattr(dep, criteria, match.group(criteria))
                break
    return dep


def _is_dev_dependency(line: str) -> bool:
    return bool(NOT_PROD_DEP.search(line))



from labels.model.file import LocationReadCloser
from labels.model.indexables import IndexedDict, IndexedList, ParsedValue
from labels.model.metadata import Artifact, Digest, HealthMetadata
from labels.model.package import Ecosystem, Package
from labels.model.package_manager import PackageManager
from labels.model.relationship import Relationship, RelationshipType
from labels.model.release import Environment
from labels.model.resolver import Resolver
from labels.parsers.cataloger.python.package_builder import new_python_package
from labels.parsers.cataloger.utils import get_enriched_location
from labels.parsers.collection import toml
from labels.utils.strings import normalize_name


def parse_uv_lock(
    _resolver: Resolver | None,
    _environment: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    file_content = toml.parse_toml_with_tree_sitter(reader.read_closer.read())

    packages = _collect_packages(file_content, reader)
    relationships = _collect_relationships(file_content, packages)

    return packages, relationships


def _collect_packages(
    file_content: IndexedDict[str, ParsedValue],
    reader: LocationReadCloser,
) -> list[Package]:
    packages = []
    toml_pkgs = file_content.get("package")
    if not isinstance(toml_pkgs, IndexedList):
        return []

    for raw_package in toml_pkgs:
        if not isinstance(raw_package, IndexedDict):
            continue

        name = str(raw_package.get("name", ""))
        version = str(raw_package.get("version", ""))

        if not version:
            continue

        new_location = (
            get_enriched_location(
                reader.location,
                line=raw_package.get_key_position("version").start.line,
                package_manager=PackageManager.UV,
            )
            if isinstance(raw_package, IndexedDict)
            else reader.location
        )

        package = new_python_package(name=name, version=version, location=new_location)
        if package:
            package.health_metadata = _extract_integrity(raw_package)
            packages.append(package)

    return packages


def _extract_hash(source: ParsedValue) -> str | None:
    if isinstance(source, IndexedDict):
        hash_value = source.get("hash")
        if isinstance(hash_value, str) and ":" in hash_value:
            return hash_value
    return None


def _extract_integrity(raw_package: IndexedDict[str, ParsedValue]) -> HealthMetadata | None:
    sdist = raw_package.get("sdist")
    selected_hash = _extract_hash(sdist)

    if not selected_hash:
        wheels = raw_package.get("wheels")
        if isinstance(wheels, IndexedList):
            for wheel in wheels:
                selected_hash = _extract_hash(wheel)
                if selected_hash:
                    break

    if not selected_hash:
        return None

    algorithm, digest_value = selected_hash.split(":", 1)
    if not algorithm or not digest_value:
        return None

    return HealthMetadata(
        artifact=Artifact(
            integrity=Digest(algorithm=algorithm, value=digest_value),
        ),
    )


def _collect_relationships(
    toml_content: IndexedDict[str, ParsedValue],
    packages: list[Package],
) -> list[Relationship]:
    relationships: list[Relationship] = []

    toml_pkgs = toml_content.get("package")
    if not isinstance(toml_pkgs, IndexedList):
        return relationships

    for package in toml_pkgs:
        dep_info = _get_dependencies(package, packages)
        if not dep_info:
            continue

        source_pkg, deps = dep_info
        if not source_pkg:
            continue

        for dep in deps:
            dep_name = _extract_dependency_name(dep)
            if not dep_name:
                continue

            dep_pkg = _find_package_by_name(packages, dep_name)
            if dep_pkg:
                relationships.append(
                    Relationship(
                        from_=dep_pkg.id_,
                        to_=source_pkg.id_,
                        type=RelationshipType.DEPENDENCY_OF_RELATIONSHIP,
                    ),
                )

    return relationships


def _get_dependencies(
    package: ParsedValue,
    packages: list[Package],
) -> tuple[Package | None, IndexedList[ParsedValue]] | None:
    if not isinstance(package, IndexedDict):
        return None

    package_name = package.get("name")
    if not isinstance(package_name, str):
        return None

    package_found = _find_package_by_name(packages, package_name)

    dependencies = package.get("dependencies")
    if not isinstance(dependencies, IndexedList):
        return None

    return package_found, dependencies


def _extract_dependency_name(dep: ParsedValue) -> str | None:
    if not isinstance(dep, IndexedDict):
        return None

    dep_name = dep.get("name")
    return str(dep_name) if isinstance(dep_name, str) else None


def _find_package_by_name(packages: list[Package], name: str) -> Package | None:
    normalized_package_name = normalize_name(name, Ecosystem.PYPI)
    return next((pkg for pkg in packages if pkg.name == normalized_package_name), None)



import itertools
import logging
import os
from collections import defaultdict
from os.path import split
from pathlib import Path

from fluidattacks_core.semver.match_versions import match_version_ranges

from labels.advisories.utils import DatabasesToDownload
from labels.enrichers.dispatcher import IMAGES_TYPES
from labels.model.file import DependencyType, Location
from labels.model.package import Package
from labels.model.relationship import Relationship, RelationshipType

LOGGER = logging.getLogger(__name__)

POSSIBLE_LOCK_FILES = {
    "Pipfile": ["Pipfile.lock"],
    "composer.json": ["composer.lock"],
    "packages.config": ["packages.lock.json"],
    "package.json": [
        "package-lock.json",
        "yarn.lock",
        "pnpm-lock.yaml",
        "bun.lock",
        # For angular libraries
        "ng-package.json",
    ],
    "Gemfile": ["Gemfile.lock", "gems.locked"],
    "Podfile": ["Podfile.lock"],
    "pyproject.toml": ["poetry.lock", "uv.lock"],
    "package.swift": ["Package.resolved"],
    "Package.swift": ["Package.resolved"],
}

LOCK_FILE_NAMES = set(itertools.chain(*POSSIBLE_LOCK_FILES.values()))


def extract_file_info(path: str) -> str:
    _, file_info = split(path)
    return file_info


def _find_lock_and_non_lock_files(
    locations: list[Location],
) -> tuple[list[Location], list[Location]]:
    non_lock_files = []
    lock_files = []

    for location in locations:
        file_info = extract_file_info(location.path())
        if file_info in POSSIBLE_LOCK_FILES:
            non_lock_files.append(location)
        elif any(file_info == lock for locks in POSSIBLE_LOCK_FILES.values() for lock in locks):
            lock_files.append(location)

    return lock_files, non_lock_files


def _get_corresponding_manifest_path(lock_file: Location) -> str | None:
    lock_path = lock_file.access_path or (
        lock_file.coordinates.real_path if lock_file.coordinates else None
    )
    if not lock_path:
        return None

    lock_file_name = extract_file_info(lock_path)
    for manifest_name, lock_names in POSSIBLE_LOCK_FILES.items():
        if lock_file_name in lock_names:
            lock_dir = parent_dir(lock_path)
            manifest_path = Path(lock_dir) / manifest_name

            if lock_file.coordinates and lock_file.coordinates.real_path:
                real_lock_dir = parent_dir(lock_file.coordinates.real_path)
                manifest_real_path = Path(real_lock_dir) / manifest_name
                if manifest_real_path.exists():
                    return str(manifest_real_path)

            if manifest_path.exists():
                return str(manifest_path)

    return None


def _find_matching_manifest(lock_file_path: str, non_lock_files: list[Location]) -> Location | None:
    for non_lock_file in non_lock_files:
        if same_directory(lock_file_path, non_lock_file.path()):
            return non_lock_file
    return None


def update_package_locations(packages: list[Package]) -> None:
    for package in packages:
        lock_files, non_lock_files = _find_lock_and_non_lock_files(package.locations)

        for lock_file in lock_files:
            lock_file_path = lock_file.access_path or (
                lock_file.coordinates.real_path if lock_file.coordinates else None
            )
            if not lock_file_path:
                continue

            matching_manifest = _find_matching_manifest(lock_file_path, non_lock_files)
            if matching_manifest:
                if lock_file.dependency_type != DependencyType.ROOT:
                    lock_file.dependency_type = matching_manifest.dependency_type
                lock_file.scope = matching_manifest.scope
                continue

            corresponding_manifest_path = _get_corresponding_manifest_path(lock_file)
            if (
                corresponding_manifest_path
                and lock_file.dependency_type == DependencyType.UNDETERMINABLE
            ):
                lock_file.dependency_type = DependencyType.TRANSITIVE


def parent_dir(path: str) -> str:
    p = path.rstrip(os.sep)
    if os.sep not in p:
        return ""
    return p.rsplit(os.sep, 1)[0]


def same_directory(lock_path: str, non_lock_path: str) -> bool:
    if lock_path.rsplit(os.sep, 1)[-1] in POSSIBLE_LOCK_FILES.get(
        non_lock_path.rsplit(os.sep, 1)[-1],
        [],
    ):
        return parent_dir(non_lock_path) == parent_dir(lock_path)

    return False


def _merge_locations(target: Package, source: Package) -> None:
    for loc in source.locations:
        if loc not in target.locations:
            target.locations.append(loc)


def obtain_locations_candidates(
    packages: dict[str, Package],
) -> tuple[
    dict[str, list[tuple[str, str, str, Location]]],
    dict[str, dict[str, list[tuple[str, str]]]],
]:
    non_lock_locations: dict[str, list[tuple[str, str, str, Location]]] = defaultdict(list)
    lock_locations: dict[str, dict[str, list[tuple[str, str]]]] = defaultdict(
        lambda: defaultdict(list)
    )
    for pkg_id, pkg in packages.items():
        for location in pkg.locations:
            if not location.access_path:
                continue
            file_name = location.access_path.rsplit(os.sep, 1)[-1]
            pkg_key = f"{pkg.name}_{pkg.language.value}_{pkg.ecosystem.value}"
            if file_name in POSSIBLE_LOCK_FILES:
                non_lock_locations[location.access_path].append(
                    (pkg_key, pkg.version, pkg_id, location)
                )
            elif file_name in LOCK_FILE_NAMES:
                lock_locations[location.access_path][pkg_key].append((pkg.version, pkg_id))

    return non_lock_locations, lock_locations


def _generate_described_by_relationships(
    lock_pkg: Package,
    non_lock_location: Location,
    searched_lock_files: list[str],
) -> list[Relationship]:
    return [
        Relationship(
            from_=f"{lock_pkg.id_}@{lock_location.location_id()}",
            to_=f"{lock_pkg.id_}@{non_lock_location.location_id()}",
            type=RelationshipType.DESCRIBED_BY_RELATIONSHIP,
        )
        for lock_location in lock_pkg.locations
        if lock_location.access_path in searched_lock_files
    ]


def merge_non_lock_locations(
    merged_packages: dict[str, Package],
) -> tuple[list[Package], list[Relationship]]:
    relationships: list[Relationship] = []
    non_lock_locations, lock_locations = obtain_locations_candidates(merged_packages)
    lock_files = list(lock_locations.keys())

    for location_path, pkgs_info in non_lock_locations.items():
        searched_lock_files = [
            lock_path
            for lock_path in lock_files
            if lock_path.rsplit(os.sep, 1)[-1]
            in POSSIBLE_LOCK_FILES.get(location_path.rsplit(os.sep, 1)[-1], [])
            and same_directory(lock_path, location_path)
        ]
        if not searched_lock_files:
            continue

        for pkg_info in pkgs_info:
            pkg_key, pkg_version, pkg_id, non_lock_location = pkg_info

            possible_equivalent_pkgs = [
                pkg_info[1]
                for searched_lock_file in searched_lock_files
                if (pkgs := lock_locations[searched_lock_file].get(pkg_key, []))
                for pkg_info in pkgs
                if match_version_ranges(pkg_info[0], pkg_version)
            ]

            if not possible_equivalent_pkgs:
                continue

            equivalent_lock_pkg_id = sorted(possible_equivalent_pkgs)[0]
            if not equivalent_lock_pkg_id:
                continue

            # Generate relationships DESCRIBED_BY: lock → non-lock
            lock_pkg = merged_packages[equivalent_lock_pkg_id]
            relationships.extend(
                _generate_described_by_relationships(
                    lock_pkg, non_lock_location, searched_lock_files
                )
            )

            merged_packages[equivalent_lock_pkg_id].locations.append(non_lock_location)
            merged_packages[pkg_id].locations.remove(non_lock_location)

            if not merged_packages[pkg_id].locations:
                del merged_packages[pkg_id]

    return list(merged_packages.values()), relationships


def merge_packages(
    packages: list[Package],
) -> tuple[list[Package], list[Relationship], DatabasesToDownload]:
    # merge locations from equal packages based on id
    same_packages: dict[str, Package] = {}
    download_images_db = False
    for pkg in packages:
        if pkg.ecosystem in IMAGES_TYPES:
            download_images_db = True

        if pkg.id_ in same_packages:
            _merge_locations(same_packages[pkg.id_], pkg)
        else:
            same_packages[pkg.id_] = pkg

    # Merge locations from lock and non-lock files and generate DESCRIBED_BY relationships
    merged_packages, described_by_relationships = merge_non_lock_locations(same_packages)

    update_package_locations(merged_packages)
    return (
        merged_packages,
        described_by_relationships,
        DatabasesToDownload(roots_db=True, images_db=download_images_db),
    )



import warnings

from fluidattacks_core.logging import init_logging, set_product_id

warnings.filterwarnings("ignore", category=ResourceWarning)
warnings.filterwarnings("ignore", category=DeprecationWarning)
warnings.filterwarnings("ignore", category=UserWarning)

set_product_id("labels")
init_logging()



import logging
from typing import NamedTuple

from packageurl import PackageURL
from pydantic import ValidationError

from labels.model.ecosystem_data.java import JavaArchive, JavaPomProject, JavaPomProperties
from labels.model.file import Location
from labels.model.package import Ecosystem, Language, Package
from labels.parsers.cataloger.java.utils.group_id_resolver import group_id_from_java_metadata
from labels.parsers.cataloger.utils import log_malformed_package_warning

LOGGER = logging.getLogger(__name__)


class JavaPackageSpec(NamedTuple):
    simple_name: str | None
    version: str | None
    location: Location
    composed_name: str | None = None
    ecosystem_data: JavaArchive | None = None


def new_java_package(package_spec: JavaPackageSpec) -> Package | None:
    simple_name = package_spec.simple_name
    version = package_spec.version

    if not simple_name or not version:
        return None

    p_url = _get_package_url_for_java(simple_name, version, package_spec.ecosystem_data)

    name = package_spec.composed_name or simple_name

    try:
        return Package(
            name=name,
            version=version,
            locations=[package_spec.location],
            language=Language.JAVA,
            ecosystem=Ecosystem.MAVEN,
            p_url=p_url,
            ecosystem_data=package_spec.ecosystem_data,
        )
    except ValidationError as ex:
        log_malformed_package_warning(package_spec.location, ex)
        return None


def new_java_package_from_maven_data(
    pom_properties: JavaPomProperties,
    parsed_pom_project: JavaPomProject | None,
    location: Location,
) -> Package | None:
    artifact_id = pom_properties.artifact_id
    version = pom_properties.version

    if not artifact_id or not version:
        return None

    ecosystem_data = JavaArchive(
        pom_properties=pom_properties,
        pom_project=parsed_pom_project if parsed_pom_project else None,
    )

    authoritative_group_id = (
        group_id_from_java_metadata(artifact_id, ecosystem_data) or pom_properties.group_id
    )

    if not authoritative_group_id:
        LOGGER.warning("No authoritative group ID found for %s in %s", artifact_id, location)
        return None

    ecosystem = Ecosystem.MAVEN
    p_url = _get_package_url_for_java(artifact_id, version, ecosystem_data)

    try:
        return Package(
            name=f"{authoritative_group_id}:{artifact_id}",
            version=version,
            locations=[location],
            ecosystem=ecosystem,
            language=Language.JAVA,
            ecosystem_data=ecosystem_data,
            p_url=p_url,
        )
    except ValidationError as ex:
        log_malformed_package_warning(location, ex)
        return None


def _get_package_url_for_java(
    name: str, version: str, ecosystem_data: JavaArchive | None = None
) -> str:
    group_id = name

    group_id_from_metadata = group_id_from_java_metadata(name, ecosystem_data)
    if group_id_from_metadata:
        group_id = group_id_from_metadata

    return PackageURL(type="maven", namespace=group_id, name=name, version=version).to_string()



from labels.model.ecosystem_data.base import EcosystemDataModel


class ApkDBEntry(EcosystemDataModel):
    package: str | None
    version: str | None
    provides: list[str]
    dependencies: list[str]
    origin_package: str | None = None
    maintainer: str | None = None
    architecture: str | None = None



from pathlib import PurePosixPath

from labels.model.file import Coordinates, Location
from labels.model.package import Language
from labels.model.package_manager import PackageManager
from labels.model.syft_sbom import SyftLocation

# Only patterns produced by syft additional catalogers inside images.
# Directory parsers already set package_manager via get_enriched_location.
_FILENAME_TO_PACKAGE_MANAGER: dict[str, PackageManager] = {
    "Podfile.lock": PackageManager.COCOAPODS,
    "Package.resolved": PackageManager.SWIFT_PM,
    ".package.resolved": PackageManager.SWIFT_PM,
    "pubspec.lock": PackageManager.PUB,
    "pubspec.yaml": PackageManager.PUB,
}


def get_package_manager(file_path: str) -> PackageManager:
    name = PurePosixPath(file_path).name
    return _FILENAME_TO_PACKAGE_MANAGER.get(name, PackageManager.UNKNOWN)


def get_language(language_str: str) -> Language:
    try:
        return Language(language_str.lower())
    except ValueError:
        return Language.UNKNOWN_LANGUAGE


def get_locations(locations: list[SyftLocation]) -> list[Location]:
    return [
        Location(
            access_path=location.access_path,
            package_manager=get_package_manager(location.access_path),
            coordinates=Coordinates(
                real_path=location.access_path,
                file_system_id=location.layer_id,
            ),
        )
        for location in locations
        if (annotation := location.annotations)
        and (evidence := annotation.evidence)
        and evidence == "primary"
    ]



import logging
from pathlib import Path

from blends.ast.build_graph import get_ast_graph_from_content
from blends.models import Content, Graph, Language, NId

LOGGER = logging.getLogger(__name__)

LANGUAGE_MAP: dict[str, Language] = {
    "python": Language.PYTHON,
    "swift": Language.SWIFT,
}


def node_str(graph: Graph, n_id: NId, key: str) -> str:
    value: str = graph.nodes[n_id][key]  # type: ignore[misc]
    return value


def maybe_node_str(graph: Graph, n_id: NId, key: str) -> str | None:
    value: str | None = graph.nodes[n_id].get(key)  # type: ignore[misc]
    return value


def node_nid(graph: Graph, n_id: NId, key: str) -> NId:
    value: NId = graph.nodes[n_id][key]  # type: ignore[misc]
    return value


def maybe_node_nid(graph: Graph, n_id: NId, key: str) -> NId | None:
    value: NId | None = graph.nodes[n_id].get(key)  # type: ignore[misc]
    return value


def parse_ast_graph(content: bytes, language: str) -> Graph | None:
    blends_language = LANGUAGE_MAP.get(language)
    if not blends_language:
        LOGGER.warning("Unable to parse content for language %s", language)
        return None

    blends_content = Content(
        content_bytes=content,
        content_str=content.decode("latin-1"),
        language=blends_language,
        path=Path(),
    )
    return get_ast_graph_from_content(blends_content)



from pydantic import BaseModel


class AwsCredentials(BaseModel):
    access_key_id: str
    secret_access_key: str
    session_token: str | None


class AwsRole(BaseModel):
    external_id: str
    role: str



from urllib3.util import parse_url

from labels.model.file import Location, LocationReadCloser
from labels.model.indexables import IndexedDict, ParsedValue
from labels.model.package import Package
from labels.model.package_manager import PackageManager
from labels.model.relationship import Relationship
from labels.model.release import Environment
from labels.model.resolver import Resolver
from labels.parsers.cataloger.dart.package_builder import DartPubspecLickEntry, new_pubspec_package
from labels.parsers.cataloger.utils import get_enriched_location
from labels.parsers.collection.yaml import parse_yaml_with_tree_sitter


def parse_pubspec_lock(
    _resolver: Resolver | None,
    _environment: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    file_content = parse_yaml_with_tree_sitter(reader.read_closer.read())
    if not isinstance(file_content, IndexedDict):
        return [], []

    yaml_packages = file_content.get("packages")
    if not isinstance(yaml_packages, IndexedDict):
        return [], []

    packages = _collect_packages(reader.location, yaml_packages)

    return packages, []


def _collect_packages(
    location: Location, yaml_packages: IndexedDict[str, ParsedValue]
) -> list[Package]:
    packages: list[Package] = []

    for package_name, package_value in yaml_packages.items():
        if not isinstance(package_value, IndexedDict):
            continue

        is_transitive = package_value.get("dependency") == "transitive"
        version = package_value.get("version")

        if not isinstance(version, str) or not package_name:
            continue

        new_location = get_enriched_location(
            location,
            line=yaml_packages.get_key_position(package_name).start.line,
            is_transitive=is_transitive,
            package_manager=PackageManager.PUB,
        )

        metadata = DartPubspecLickEntry(
            hosted_url=_get_hosted_url(package_value),
            vcs_url=_get_vcs_url(package_value),
        )

        package = new_pubspec_package(package_name, version, new_location, metadata)
        if package:
            packages.append(package)

    return packages


def _get_hosted_url(entry: IndexedDict[str, ParsedValue]) -> str:
    hosted = entry.get("hosted")
    description = entry.get("description")

    if hosted != "hosted" or not isinstance(description, IndexedDict):
        return ""

    description_url = description.get("url")
    if not isinstance(description_url, str) or description_url == "https://pub.dartlang.org":
        return ""

    host_from_url = parse_url(description_url).host

    return host_from_url if host_from_url else description_url


def _get_vcs_url(entry: IndexedDict[str, ParsedValue]) -> str:
    source = entry.get("source")
    description = entry.get("description")

    if source != "git" or not isinstance(description, IndexedDict):
        return ""

    url = description.get("url")
    resolved_ref = description.get("resolved-ref")
    path = description.get("path")

    if not url or not resolved_ref:
        return ""

    if path == "." or not path:
        return f"{url}@{resolved_ref}"

    return f"{url}@{resolved_ref}#{path}"



import atexit
import logging
import sqlite3
from pathlib import Path
from typing import Literal

import boto3
import zstd
from botocore import UNSIGNED
from botocore.config import Config
from platformdirs import user_data_dir

LOGGER = logging.getLogger(__name__)


class BaseDatabase:
    BUCKET_NAME = "skims.sca"
    S3_SERVICE_NAME: Literal["s3"] = "s3"
    CONFIG_DIRECTORY = user_data_dir(
        appname="fluid-labels",
        appauthor="fluidattacks",
        ensure_exists=True,
    )

    def __init__(self, db_name: str) -> None:
        self.db_name = db_name
        self.bucket_file_key = f"{db_name}.zst"
        self.db_local_path = str(Path(self.CONFIG_DIRECTORY, db_name))
        self.db_local_compressed_path = f"{self.db_local_path}.zst"
        self.connection: sqlite3.Connection | None = None
        self.s3_client = boto3.client(
            service_name=self.S3_SERVICE_NAME,
            config=Config(
                region_name="us-east-1",
                signature_version=UNSIGNED,  # type: ignore[misc]
            ),
        )

    def _get_database_file(self) -> None:
        LOGGER.info("⬇️ Downloading advisories database")
        self.s3_client.download_file(
            Bucket=self.BUCKET_NAME,
            Key=self.bucket_file_key,
            Filename=self.db_local_compressed_path,
        )
        LOGGER.info("🗜️ Decompressing advisories database")

        try:
            with Path(self.db_local_compressed_path).open("rb") as compressed_file:
                compressed_data: bytes = compressed_file.read()
            uncompressed_data: bytes = zstd.decompress(compressed_data)
            with Path(self.db_local_path).open("wb") as output_file:
                output_file.write(uncompressed_data)
        except Exception:
            LOGGER.exception("❌ Unable to decompress database %s", self.db_name)

    def _initialize_db(self) -> bool:
        local_database_exists = Path(self.db_local_path).is_file()

        try:
            if self.is_up_to_date(local_database_exists=local_database_exists):
                LOGGER.info("✅ Advisories database is up to date")
                return True
            self._get_database_file()
            Path(self.db_local_compressed_path).unlink()
        except Exception:
            if local_database_exists:
                LOGGER.warning(
                    "⚠️ Advisories may be outdated, unable to update database",
                )
                return True

            LOGGER.exception(
                "❌ Advisories won't be included, unable to download database",
            )
            return False
        else:
            return True

    def is_up_to_date(self, *, local_database_exists: bool) -> bool:
        db_metadata = self.s3_client.head_object(
            Bucket=self.BUCKET_NAME,
            Key=self.bucket_file_key,
        )
        return (
            local_database_exists
            and Path(self.db_local_path).stat().st_mtime >= db_metadata["LastModified"].timestamp()
        )

    def initialize(self) -> None:
        if self.connection is None and self._initialize_db():
            self.connection = sqlite3.connect(
                self.db_local_path,
                check_same_thread=False,
            )
            atexit.register(self.connection.close)

    def get_connection(self) -> sqlite3.Connection:
        if self.connection is not None:
            return self.connection
        self.connection = sqlite3.connect(
            self.db_local_path,
            check_same_thread=False,
        )
        atexit.register(self.connection.close)
        return self.connection



from blends.models import Graph, NId
from blends.query import adj
from blends.utilities.text_nodes import node_to_str

from labels.model.file import LocationReadCloser
from labels.model.package import Package
from labels.model.package_manager import PackageManager
from labels.model.relationship import Relationship
from labels.model.release import Environment
from labels.model.resolver import Resolver
from labels.parsers.cataloger.graph_parser import (
    maybe_node_nid,
    maybe_node_str,
    node_nid,
    node_str,
    parse_ast_graph,
)
from labels.parsers.cataloger.swift.package_builder import new_swift_package_manager_package
from labels.parsers.cataloger.utils import get_enriched_location


def parse_package_swift(
    _resolver: Resolver | None,
    _environment: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    file_content = reader.read_closer.read().encode("utf-8")
    if not file_content or not (graph := parse_ast_graph(file_content, "swift")):
        return [], []

    dependencies = _parse_dependencies(graph)
    if dependencies is None:
        return [], []

    packages = _collect_packages(dependencies, graph, reader)

    return packages, []


def _parse_dependencies(graph: Graph) -> NId | None:
    dependencies = None

    for node in graph.nodes:
        if (
            maybe_node_str(graph, node, "label_type") == "call_expression"
            and (children := adj(graph, node))
            and len(children) == 2
            and maybe_node_str(graph, children[0], "label_text") == "Package"
        ):
            for child in adj(graph, children[1], depth=2):
                name_id = maybe_node_nid(graph, child, "label_field_name")
                if (
                    maybe_node_str(graph, child, "label_type") == "value_argument"
                    and name_id is not None
                    and (name_value_id := adj(graph, name_id)[0])
                    and maybe_node_str(graph, name_value_id, "label_text") == "dependencies"
                ):
                    dependencies = node_nid(graph, child, "label_field_value")
                    break

    return dependencies


def _collect_packages(
    dep_node: NId,
    graph: Graph,
    reader: LocationReadCloser,
) -> list[Package]:
    packages: list[Package] = []

    for package_node in adj(graph, dep_node, depth=2):
        if _is_valid_package_node(graph, package_node):
            source_url, package_version = _extract_package_details(graph, package_node)

            new_location = get_enriched_location(
                reader.location,
                line=int(node_str(graph, package_node, "label_l")),
                package_manager=PackageManager.SWIFT_PM,
            )

            package = new_swift_package_manager_package(
                source_url=source_url, version=package_version, location=new_location
            )

            if package:
                packages.append(package)

    return packages


def _is_valid_package_node(graph: Graph, package_node: NId) -> bool:
    return bool(
        node_str(graph, package_node, "label_type") == "call_expression"
        and (child_value := adj(graph, package_node)[0])
        and node_to_str(graph, child_value) != ".package(",
    )


def _extract_package_details(
    graph: Graph,
    package_node: NId,
) -> tuple[str | None, str | None]:
    package_version = None
    source_url = None

    for argument_node in adj(graph, package_node, 3):
        name_id = maybe_node_nid(graph, argument_node, "label_field_name")
        if (
            maybe_node_str(graph, argument_node, "label_type") != "value_argument"
            or name_id is None
        ):
            continue

        argument_name = node_to_str(graph, name_id)
        value_id = maybe_node_nid(graph, argument_node, "label_field_value")
        if value_id is None:
            continue

        if argument_name == "url":
            source_url = node_to_str(graph, value_id)
        elif argument_name in {"from", "version"}:
            package_version = node_to_str(graph, value_id).replace('"', "").replace("'", "")

    return source_url, package_version



import logging
import os
import time
from concurrent.futures import ThreadPoolExecutor

from labels.advisories.images import DATABASE as IMAGES_DATABASE
from labels.advisories.roots import DATABASE as ROOTS_DATABASE
from labels.advisories.utils import DatabasesToDownload
from labels.config.utils import guess_environment
from labels.core.merge_packages import merge_packages
from labels.core.source_dispatcher import resolve_sbom_source
from labels.core.top_parents import calculate_top_parents_for_packages
from labels.domain.cloudwatch import process_sbom_metrics
from labels.domain.tracks import send_event_to_tracks
from labels.enrichers.dispatcher import (
    complete_package_advisories,
)
from labels.model.core import SbomConfig
from labels.model.package import Package
from labels.model.relationship import Relationship
from labels.model.syft_sbom import SyftSBOM
from labels.output.dispatcher import dispatch_sbom_output
from labels.parsers.operations.package_operation import (
    package_operations_factory,
    package_operations_from_sbom,
)
from labels.resolvers.directory import Directory
from labels.utils.tracks import count_vulns_by_severity

LOGGER = logging.getLogger(__name__)


def configure_logging(sbom_config: SbomConfig) -> None:
    if sbom_config.debug:
        logger = logging.getLogger()
        logger.setLevel(logging.DEBUG)


def initialize_advisories_databases(dbs_to_download: DatabasesToDownload) -> None:
    LOGGER.info("📦 Initializing advisories databases")
    if dbs_to_download.roots_db:
        ROOTS_DATABASE.initialize()
    if dbs_to_download.images_db:
        IMAGES_DATABASE.initialize()


def gather_packages_and_relationships(
    resolver: Directory | SyftSBOM,
    max_workers: int = 32,
    *,
    include_package_advisories: bool = True,
) -> tuple[list[Package], list[Relationship]]:
    LOGGER.info("📦 Gathering packages and relationships")
    packages, relationships = (
        package_operations_factory(resolver)
        if isinstance(resolver, Directory)
        else package_operations_from_sbom(resolver)
    )
    merged_packages, described_by_relationships, dbs_to_download = merge_packages(packages)
    LOGGER.debug("PACKAGES: %d - MERGED: %d", len(packages), len(merged_packages))
    LOGGER.debug(
        "RELATIONSHIPS: %d - DESCRIBED BY: %d",
        len(relationships),
        len(described_by_relationships),
    )

    all_relationships = relationships + described_by_relationships

    if include_package_advisories:
        initialize_advisories_databases(dbs_to_download)

    worker_count = min(
        max_workers,
        (os.cpu_count() or 1) * 5 if os.cpu_count() is not None else max_workers,
    )
    with ThreadPoolExecutor(max_workers=worker_count) as executor:
        LOGGER.info("📦 Gathering additional package information")
        if include_package_advisories:
            packages = list(executor.map(complete_package_advisories, merged_packages))
        else:
            packages = merged_packages

    LOGGER.info("📦 Calculating top parents")
    packages = calculate_top_parents_for_packages(packages, all_relationships)

    LOGGER.info("✅ Found %d packages", len(packages))

    return packages, all_relationships


def execute_labels_scan(sbom_config: SbomConfig) -> None:
    try:
        configure_logging(sbom_config)

        LOGGER.info("🚀 Starting Labels SBOM scan")
        main_sbom_resolver = resolve_sbom_source(sbom_config)

        LOGGER.info(
            "📦 Generating SBOM from %s: %s", sbom_config.source_type.value, sbom_config.source
        )
        start_time = time.perf_counter()
        packages, relationships = gather_packages_and_relationships(
            main_sbom_resolver,
            include_package_advisories=sbom_config.include_package_advisories,
        )
        end_time = time.perf_counter() - start_time
        process_sbom_metrics(sbom_config.execution_id, end_time, sbom_config.source_type)

        LOGGER.info("📦 Preparing %s report", sbom_config.output_format.value)
        dispatch_sbom_output(
            packages=packages,
            relationships=relationships,
            config=sbom_config,
            resolver=main_sbom_resolver,
        )
        send_event_to_tracks(
            sbom_config=sbom_config,
            packages_amount=len(packages),
            relationships_amount=len(relationships),
            vulns_summary=count_vulns_by_severity(packages),
        )
    except Exception:
        if guess_environment() == "production":
            LOGGER.exception(
                "Error executing labels scan. Output SBOM was not generated.",
                extra={"execution_id": sbom_config.execution_id},
            )
            return
        raise



from packageurl import PackageURL
from pydantic import ValidationError

from labels.model.file import Location
from labels.model.package import Ecosystem, Language, Package
from labels.parsers.cataloger.utils import log_malformed_package_warning


def new_go_package(name: str | None, version: str | None, location: Location) -> Package | None:
    if not name or not version:
        return None

    p_url = _get_package_url_for_go(name, version)

    try:
        return Package(
            name=name,
            version=version,
            ecosystem=Ecosystem.GO,
            locations=[location],
            p_url=p_url,
            ecosystem_data=None,
            language=Language.GO,
        )
    except ValidationError as ex:
        log_malformed_package_warning(location, ex)
        return None


def _get_package_url_for_go(module_name: str, module_version: str) -> str:
    fields = module_name.split("/")

    namespace = ""
    name = ""
    subpath = ""

    if len(fields) == 1:
        name = fields[0]
    elif len(fields) == 2:
        name = fields[1]
        namespace = fields[0]
    else:
        name = fields[2]
        namespace = "/".join(fields[:2])
        subpath = "/".join(fields[3:])

    return PackageURL(
        type="golang", namespace=namespace, name=name, version=module_version, subpath=subpath
    ).to_string()



import json
import logging
from datetime import UTC, datetime
from pathlib import Path

from jsonschema import validate
from jsonschema.exceptions import ValidationError

from labels.model.core import SbomConfig
from labels.model.package import Package
from labels.model.relationship import Relationship
from labels.model.resolver import Resolver
from labels.model.syft_sbom import SyftSBOM
from labels.output.fluid.file_builder import (
    build_relationship_map,
    build_sbom_metadata,
    serialize_packages,
)
from labels.output.fluid.schema import FLUID_SBOM_JSON_SCHEMA
from labels.output.fluid.types import FluidSBOM
from labels.output.utils import set_namespace_version
from labels.utils.exceptions import FluidJSONValidationError

LOGGER = logging.getLogger(__name__)


def write_json_to_file(path: Path, content: FluidSBOM) -> None:
    with path.open("w", encoding="utf-8") as f:
        json.dump(content, f, indent=4)
    LOGGER.info("✅ Output file successfully generated")


def validate_fluid_json(result: FluidSBOM, file_path: Path) -> None:
    try:
        validate(result, FLUID_SBOM_JSON_SCHEMA)

    except ValidationError as ex:
        raise FluidJSONValidationError(ex) from None

    LOGGER.info("🆗 Valid Fluid JSON format, generating output file at %s", file_path)


def format_fluid_sbom(
    *,
    packages: list[Package],
    relationships: list[Relationship],
    config: SbomConfig,
    resolver: Resolver | SyftSBOM,
) -> None:
    now_utc = datetime.now(UTC).isoformat()
    namespace, version = set_namespace_version(config=config, resolver=resolver)
    file_path = Path(f"{config.output}.json")

    sbom_pkgs = serialize_packages(packages)
    sbom_relationships = build_relationship_map(relationships)
    sbom_metadata = build_sbom_metadata(namespace, version, now_utc)

    result = FluidSBOM(
        sbom_details=sbom_metadata,
        packages=sbom_pkgs,
        relationships=sbom_relationships,
    )

    validate_fluid_json(result, file_path)
    write_json_to_file(file_path, result)



import re
from typing import NamedTuple

from labels.model.file import Location, LocationReadCloser
from labels.model.indexables import IndexedDict, ParsedValue
from labels.model.package import Package
from labels.model.package_manager import PackageManager
from labels.model.relationship import Relationship
from labels.model.release import Environment
from labels.model.resolver import Resolver
from labels.parsers.cataloger.javascript.common import (
    DependencyMap,
    add_dependency_to_root_relationship,
    build_relationships,
    create_root_package_from_package_json,
    find_root_location,
    get_direct_dependencies_from_package_json,
    group_packages_by_name,
)
from labels.parsers.cataloger.javascript.package_builder import new_simple_npm_package
from labels.parsers.cataloger.utils import get_enriched_location, parse_integrity_to_health_metadata
from labels.parsers.collection.yaml import parse_yaml_with_tree_sitter

VERSION_PATTERN = re.compile(r"(\d+\.\d+\.\d+(-[0-9A-Za-z\.]+)?)")


class PnpmPackageCreationDetails(NamedTuple):
    package_key: str
    package_spec: IndexedDict[str, ParsedValue]
    packages_object: IndexedDict[str, ParsedValue]
    direct_dependencies: set[str]
    base_location: Location


class PnpmPackageInfo(NamedTuple):
    name: str
    key: str
    is_dev: bool
    dependencies: set[str]


def parse_pnpm_lock(
    _resolver: Resolver | None,
    _environment: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    content = reader.read_closer.read()
    file_content = parse_yaml_with_tree_sitter(content)
    if not isinstance(file_content, IndexedDict):
        return [], []

    direct_dependencies = get_direct_dependencies_from_package_json(reader.location)
    root_package = create_root_package_from_package_json(
        reader.location, package_manager=PackageManager.PNPM
    )

    packages = _collect_packages(reader, file_content, direct_dependencies)

    if root_package:
        packages.append(root_package)

    dependency_map = _build_dependency_map_pnpm(file_content)
    key_map = _build_key_map_pnpm(file_content)

    relationships = build_relationships(packages, dependency_map, reader.location, key_map)

    if root_package:
        root_relationships = _create_root_relationships(
            root_package, packages, direct_dependencies, reader.location
        )
        relationships.extend(root_relationships)

    return packages, relationships


def _create_root_relationships(
    root_package: Package,
    packages: list[Package],
    direct_dependencies: set[str],
    location: Location,
) -> list[Relationship]:
    """Create relationships from direct dependencies to root package."""
    relationships: list[Relationship] = []

    if not (location.coordinates and location.coordinates.real_path):
        return relationships

    lockfile_path = location.coordinates.real_path
    root_location = find_root_location(root_package, lockfile_path)

    if not root_location:
        return relationships

    packages_by_name = group_packages_by_name(packages, root_package.id_)

    for dep_name in direct_dependencies:
        for dep_pkg in packages_by_name.get(dep_name, []):
            add_dependency_to_root_relationship(
                dep_pkg, root_package, root_location, lockfile_path, relationships
            )

    return relationships


def _collect_packages(
    reader: LocationReadCloser,
    file_content: IndexedDict[str, ParsedValue],
    direct_dependencies: set[str],
) -> list[Package]:
    packages: list[Package] = []

    packages_object = file_content.get("packages")
    if not isinstance(packages_object, IndexedDict):
        return packages

    for package_key, pkg_spec in packages_object.items():
        if not isinstance(pkg_spec, IndexedDict):
            continue

        package_creation_details = PnpmPackageCreationDetails(
            package_key=package_key,
            package_spec=pkg_spec,
            packages_object=packages_object,
            direct_dependencies=direct_dependencies,
            base_location=reader.location,
        )

        package = _process_package(package_creation_details)
        if package:
            packages.append(package)

    return packages


def _process_package(creation_details: PnpmPackageCreationDetails) -> Package | None:
    name_and_version = _parse_package_key(
        creation_details.package_key, creation_details.package_spec
    )
    if name_and_version is None:
        return None

    package_name, package_version = name_and_version

    is_dev = creation_details.package_spec.get("dev") is True

    package_info = PnpmPackageInfo(
        name=package_name,
        key=creation_details.package_key,
        dependencies=creation_details.direct_dependencies,
        is_dev=is_dev,
    )

    position = creation_details.packages_object.get_key_position(package_info.key)
    is_transitive = package_info.name not in package_info.dependencies

    new_location = get_enriched_location(
        creation_details.base_location,
        line=position.start.line,
        is_transitive=is_transitive,
        is_dev=package_info.is_dev,
        package_manager=PackageManager.PNPM,
    )

    resolution = creation_details.package_spec.get("resolution")
    integrity_str = resolution.get("integrity") if isinstance(resolution, IndexedDict) else None
    health_metadata = parse_integrity_to_health_metadata(
        str(integrity_str) if integrity_str else None
    )

    return new_simple_npm_package(new_location, package_name, package_version, health_metadata)


def _parse_package_key(package: str, spec: IndexedDict[str, ParsedValue]) -> tuple[str, str] | None:
    if package.startswith("github"):
        package_name = spec.get("name")
        package_version = spec.get("version")
    else:
        package_info: list[str] = VERSION_PATTERN.split(package.strip("\"'"))
        if len(package_info) < 2:
            return None
        package_name = package_info[0].lstrip("/")[0:-1]
        package_version = package_info[1]

    if not isinstance(package_name, str) or not isinstance(package_version, str):
        return None

    return package_name, package_version


def _process_package_for_dependency_map(
    package_key: str,
    package_value: IndexedDict[str, ParsedValue],
    package_name_regex: re.Pattern[str],
    dependency_map: DependencyMap,
) -> None:
    cleaned_key = package_name_regex.sub(r"\1", package_key)
    name_and_version = _parse_package_key(cleaned_key, package_value)
    if not name_and_version:
        return

    package_name, package_version = name_and_version
    map_key = f"{package_name}@{package_version}"

    if map_key in dependency_map:
        return

    deps = package_value.get("dependencies")
    peer_deps = package_value.get("peerDependencies")
    opt_deps = package_value.get("optionalDependencies")

    dependencies_dict: dict[str, ParsedValue] = {}
    if isinstance(deps, IndexedDict):
        dependencies_dict.update(deps)
    if isinstance(peer_deps, IndexedDict):
        dependencies_dict.update(peer_deps)
    if isinstance(opt_deps, IndexedDict):
        dependencies_dict.update(opt_deps)

    if not dependencies_dict:
        return

    deps_dict = _extract_dependency_requirements(dependencies_dict)
    if deps_dict:
        dependency_map[map_key] = deps_dict


def _build_key_map_pnpm(
    package_yaml: IndexedDict[str, ParsedValue],
) -> dict[tuple[str, int], str]:
    """Build a map from (package_name, line) to package_key (name@version)."""
    key_map: dict[tuple[str, int], str] = {}

    snapshots = package_yaml.get("snapshots")
    packages_items = package_yaml.get("packages")

    sources_to_check = []
    if isinstance(snapshots, IndexedDict):
        sources_to_check.append(snapshots)
    if isinstance(packages_items, IndexedDict):
        sources_to_check.append(packages_items)

    package_name_regex = re.compile(r"^/?([^(]*)(?:\(.*\))*$")

    for source_items in sources_to_check:
        for package_key, package_value in source_items.items():
            if not isinstance(package_value, IndexedDict):
                continue

            cleaned_key = package_name_regex.sub(r"\1", package_key)
            name_and_version = _parse_package_key(cleaned_key, package_value)
            if not name_and_version:
                continue

            package_name, package_version = name_and_version
            map_key = f"{package_name}@{package_version}"

            position = source_items.get_key_position(package_key)
            key_map[(package_name, position.start.line)] = map_key

    return key_map


def _build_dependency_map_pnpm(
    package_yaml: IndexedDict[str, ParsedValue],
) -> DependencyMap:
    dependency_map: DependencyMap = {}

    snapshots = package_yaml.get("snapshots")
    packages_items = package_yaml.get("packages")

    sources_to_check = []
    if isinstance(snapshots, IndexedDict):
        sources_to_check.append(snapshots)
    if isinstance(packages_items, IndexedDict):
        sources_to_check.append(packages_items)

    package_name_regex = re.compile(r"^/?([^(]*)(?:\(.*\))*$")

    for source_items in sources_to_check:
        for package_key, package_value in source_items.items():
            if isinstance(package_value, IndexedDict):
                _process_package_for_dependency_map(
                    package_key, package_value, package_name_regex, dependency_map
                )

    return dependency_map


def _extract_dependency_requirements(
    dependencies: dict[str, ParsedValue],
) -> dict[str, str]:
    deps_dict: dict[str, str] = {}

    for raw_dep_name, raw_dep_version in dependencies.items():
        if not isinstance(raw_dep_version, str):
            continue

        # Extract package name (handles scoped packages like @babel/core)
        dep_name = _extract_package_name_from_key_dependency(raw_dep_name)
        if not dep_name:
            # If extraction fails, use the raw name (for simple packages)
            dep_name = raw_dep_name
        dep_version = _extract_version_from_value_dependency(raw_dep_version)

        # Store the version requirement
        deps_dict[dep_name] = dep_version

    return deps_dict


def _extract_package_name_from_key_dependency(item: str) -> str | None:
    # Regex pattern to extract the package name
    pattern = r"^@?[\w-]+/[\w-]+$"
    match = re.match(pattern, item)
    if match:
        return match.group(0)
    return None


def _extract_version_from_value_dependency(item: str) -> str:
    # Regex pattern to extract the version number before any parentheses
    # Includes pre-release identifiers like -beta.2, -alpha.1, -rc.3, etc.
    pattern = r"^(\d+\.\d+\.\d+(?:-[\w.]+)?)"
    match = re.match(pattern, item)
    if match:
        return match.group(1)
    return item



from fluidattacks_core.semver.match_versions import (
    _simplify_pre_release,
    compare_pre_releases,
    compare_versions,
    match_vulnerable_versions,
    normalize,
)
from pydantic import BaseModel

from labels.model.advisories import (
    Advisory,
    AdvisoryFixMetadata,
    AdvisoryFixMetadataNew,
    FixMetadata,
    ResidualCVES,
    SeverityCVES,
    UpgradeType,
)
from labels.model.package import Ecosystem
from labels.utils.version_stability import is_unstable_version

COMMON_QUALIFIERS = {
    "RELEASE",
    "Final",
    "GA",
    "GA.RELEASE",
    "RELEASE.FINAL",
    "FINAL",
    "STABLE",
    "LTS",
    "GA.FINAL",
}


class VersionMetadata(BaseModel):
    pre: str | None
    qual: str | None
    build: str | None


def _advisory_key(advisory: Advisory) -> tuple[str, str]:
    """Generate a hashable key for an advisory for fast set operations."""
    return (advisory.id, advisory.vulnerable_version)


def _is_qualifier(part: str) -> bool:
    return part.upper() in {q.upper() for q in COMMON_QUALIFIERS}


def _extract_build_metadata(parts: list[str]) -> tuple[list[str], str | None]:
    clean_parts = []
    build_metadata = None

    for part in parts:
        if "+" in part:
            num_part, build_part = part.split("+", 1)
            clean_parts.append(num_part)
            build_metadata = build_part
        else:
            clean_parts.append(part)

    return clean_parts, build_metadata


def _extract_qualifier(parts: list[str]) -> tuple[list[str], str | None]:
    for index_part in range(1, len(parts)):
        if _is_qualifier(parts[index_part]):
            return parts[:index_part] + parts[index_part + 1 :], parts[index_part]
    return parts, None


def _normalize_version_for_comparison(
    version: str,
) -> tuple[list[str], str | None, str | None, str | None]:
    parts, pre_release = normalize(version)

    clean_parts, build_metadata = _extract_build_metadata(parts)
    final_parts, qualifier = _extract_qualifier(clean_parts)

    simplified_pre_release = None
    if pre_release:
        simplified_pre_release = _simplify_pre_release(pre_release.lower())

    return final_parts, simplified_pre_release, qualifier, build_metadata


def _are_versions_equivalent(version1: str, version2: str) -> bool:
    v1_parts, v1_pre, _, _ = _normalize_version_for_comparison(version1)
    v2_parts, v2_pre, _, _ = _normalize_version_for_comparison(version2)
    return v1_parts == v2_parts and v1_pre == v2_pre


def _is_build_metadata_greater(v_build: str | None, dep_build: str | None) -> bool | None:
    if v_build is None and dep_build is not None:
        return True
    if v_build is not None and dep_build is None:
        return False
    if v_build is not None and dep_build is not None:
        return v_build > dep_build
    return None


def _compare_single_part(v_part: str, dep_part: str) -> bool | None:
    if v_part.isdigit() and dep_part.isdigit():
        v_num = int(v_part)
        dep_num = int(dep_part)
        return v_num > dep_num if v_num != dep_num else None
    return v_part > dep_part if v_part != dep_part else None


def _compare_common_parts(v_parts: list[str], dep_parts: list[str]) -> bool | None:
    min_parts = min(len(v_parts), len(dep_parts))

    for index_part in range(min_parts):
        comparison = _compare_single_part(v_parts[index_part], dep_parts[index_part])
        if comparison is not None:
            return comparison

    return None


def _compare_equal_version_parts(
    v_pre: str | None,
    dep_pre: str | None,
    v_build: str | None,
    dep_build: str | None,
) -> bool | None:
    if v_pre is None and dep_pre is not None:
        return True
    if v_pre is not None and dep_pre is not None:
        return compare_pre_releases(v_pre, dep_pre)
    return _is_build_metadata_greater(v_build, dep_build)


def _is_version_greater_than_dep(version: str, dep_version: str) -> bool:
    if not version or not dep_version or "${" in version or "${" in dep_version:
        return False

    if _are_versions_equivalent(version, dep_version):
        return False

    if compare_versions(version1=version, version2=dep_version, include_same=False):
        return True

    dep_parts, dep_pre, _, dep_build = _normalize_version_for_comparison(dep_version)
    v_parts, v_pre, _, v_build = _normalize_version_for_comparison(version)

    if v_parts == dep_parts:
        result = _compare_equal_version_parts(v_pre, dep_pre, v_build, dep_build)
        if result is not None:
            return result

    common_parts_comparison = _compare_common_parts(v_parts, dep_parts)
    if common_parts_comparison is not None:
        return common_parts_comparison

    return len(v_parts) > len(dep_parts)


def _get_upgrade_type_from_numeric_parts(
    cur_nums: list[int], tgt_nums: list[int]
) -> UpgradeType | None:
    if tgt_nums[0] > cur_nums[0]:
        return UpgradeType.MAJOR
    if tgt_nums[1] > cur_nums[1]:
        return UpgradeType.MINOR
    if tgt_nums[2] > cur_nums[2]:
        return UpgradeType.PATCH
    return None


def _get_upgrade_type_from_metadata(
    cur_meta: VersionMetadata, tgt_meta: VersionMetadata
) -> UpgradeType:
    if tgt_meta.pre and cur_meta.pre != tgt_meta.pre:
        return UpgradeType.PATCH
    if tgt_meta.pre is None and cur_meta.pre is not None:
        return UpgradeType.PATCH
    if cur_meta.qual != tgt_meta.qual or cur_meta.build != tgt_meta.build:
        return UpgradeType.PATCH
    return UpgradeType.UNKNOWN


def _get_upgrade_type_from_index(part_index: int) -> UpgradeType:
    if part_index == 0:
        return UpgradeType.MAJOR
    if part_index == 1:
        return UpgradeType.MINOR
    return UpgradeType.PATCH


def _get_upgrade_type_from_parts(cur_parts: list[str], tgt_parts: list[str]) -> UpgradeType | None:
    min_parts = min(len(cur_parts), len(tgt_parts))

    for part_index in range(min_parts):
        if cur_parts[part_index] == tgt_parts[part_index]:
            continue

        if cur_parts[part_index].isdigit() and tgt_parts[part_index].isdigit():
            cur_num = int(cur_parts[part_index])
            tgt_num = int(tgt_parts[part_index])
            if cur_num != tgt_num:
                return _get_upgrade_type_from_index(part_index)

        return UpgradeType.PATCH

    return UpgradeType.PATCH if len(cur_parts) != len(tgt_parts) else None


def get_upgrade_type(current: str, target: str) -> UpgradeType:
    cur_parts, cur_pre, cur_qual, cur_build = _normalize_version_for_comparison(current)
    tgt_parts, tgt_pre, tgt_qual, tgt_build = _normalize_version_for_comparison(target)

    cur_nums = [int(p) if p.isdigit() else 0 for p in ([*cur_parts, "0", "0", "0"])[:3]]
    tgt_nums = [int(p) if p.isdigit() else 0 for p in ([*tgt_parts, "0", "0", "0"])[:3]]

    numeric_upgrade = _get_upgrade_type_from_numeric_parts(cur_nums, tgt_nums)
    if numeric_upgrade:
        return numeric_upgrade

    part_upgrade = _get_upgrade_type_from_parts(cur_parts, tgt_parts)
    if part_upgrade:
        return part_upgrade

    cur_meta = VersionMetadata(pre=cur_pre, qual=cur_qual, build=cur_build)
    tgt_meta = VersionMetadata(pre=tgt_pre, qual=tgt_qual, build=tgt_build)
    return _get_upgrade_type_from_metadata(cur_meta, tgt_meta)


def version_sort_key(version: str) -> tuple[object, ...]:
    parts, pre, _, _ = _normalize_version_for_comparison(version)
    nums = tuple(int(p) if p.isdigit() else 0 for p in ([*parts, "0", "0", "0"])[:3])
    is_release = 1 if pre is None else 0
    pre_str = pre or ""
    return (*nums, is_release, pre_str)


def _filter_stable_versions(versions: list[str], ecosystem_str: str) -> list[str]:
    eco = Ecosystem(ecosystem_str) if ecosystem_str in Ecosystem.__members__.values() else None
    return [v for v in versions if not is_unstable_version(v, eco)]


def match_fixed_versions(
    dep_version: str, advisory: Advisory, safe_versions: list[str] | None
) -> Advisory:
    cve_fix_versions = advisory.fixed_versions or safe_versions
    closest_fix = None
    upgrade_type = UpgradeType.UNKNOWN
    breaking_change = False

    if cve_fix_versions:
        stable = _filter_stable_versions(cve_fix_versions, advisory.ecosystem)
        fixed = [v for v in stable if _is_version_greater_than_dep(v, dep_version)]
        if fixed:
            closest_fix = min(fixed, key=version_sort_key)
            upgrade_type = get_upgrade_type(dep_version, closest_fix)
            breaking_change = upgrade_type == UpgradeType.MAJOR

    closest_safe = None
    if safe_versions:
        stable_safe = _filter_stable_versions(safe_versions, advisory.ecosystem)
        safe = [v for v in stable_safe if _is_version_greater_than_dep(v, dep_version)]
        if safe:
            closest_safe = min(safe, key=version_sort_key)

    if not closest_fix and closest_safe:
        closest_fix = closest_safe
        upgrade_type = get_upgrade_type(dep_version, closest_fix)
        breaking_change = upgrade_type == UpgradeType.MAJOR

    if closest_fix is None:
        advisory.fix_metadata = None
        return advisory

    advisory.fix_metadata = AdvisoryFixMetadata(
        closest_fix_version=closest_fix,
        upgrade_type=upgrade_type,
        breaking_change=breaking_change,
        closest_safe_version=closest_safe,
    )

    return advisory


def _create_severity_cves(advisories: list[Advisory]) -> SeverityCVES | None:
    total = len(advisories)
    if total == 0:
        return None
    critical = []
    high = []
    medium = []
    low = []

    for advisory in advisories:
        severity_lower = advisory.severity_level.lower() if advisory.severity_level else "low"
        if severity_lower == "critical":
            critical.append(advisory.id)
        elif severity_lower == "high":
            high.append(advisory.id)
        elif severity_lower == "medium":
            medium.append(advisory.id)
        elif severity_lower == "low":
            low.append(advisory.id)

    return SeverityCVES(
        critical=critical if critical else None,
        high=high if high else None,
        medium=medium if medium else None,
        low=low if low else None,
        total=len(advisories),
    )


def _get_residual_cves(
    candidate_fix_version: str,
    dep_version: str,
    all_package_advisories: list[Advisory],
    base_advisories: list[Advisory] | None = None,
    base_keys: set[tuple[str, str]] | None = None,
) -> ResidualCVES | None:
    # Compute base_advisories only if not provided
    if base_advisories is None:
        base_advisories = [
            a
            for a in all_package_advisories
            if match_vulnerable_versions(dep_version, a.vulnerable_version)
        ]
    if base_keys is None:
        base_keys = {_advisory_key(a) for a in base_advisories}

    candidate_advisories = [
        advisory
        for advisory in all_package_advisories
        if match_vulnerable_versions(candidate_fix_version, advisory.vulnerable_version)
    ]

    candidate_keys = {_advisory_key(a) for a in candidate_advisories}

    remedied_cves = [a for a in base_advisories if _advisory_key(a) not in candidate_keys]
    maintained_cves = [a for a in candidate_advisories if _advisory_key(a) in base_keys]
    introduced_cves = [a for a in candidate_advisories if _advisory_key(a) not in base_keys]

    remedied = _create_severity_cves(remedied_cves)
    maintained = _create_severity_cves(maintained_cves)
    introduced = _create_severity_cves(introduced_cves)

    if remedied is None and maintained is None and introduced is None:
        return None

    return ResidualCVES(
        remedied=remedied,
        maintained=maintained,
        introduced=introduced,
    )


def _calculate_minimal_fix(  # noqa: PLR0913
    dep_version: str,
    advisories: list[Advisory],
    cve_fix_versions: list[str] | None,
    base_advisories: list[Advisory],
    base_keys: set[tuple[str, str]],
    ecosystem_str: str = "",
) -> FixMetadata | None:
    if cve_fix_versions:
        stable = _filter_stable_versions(cve_fix_versions, ecosystem_str)
        fixed = [v for v in stable if _is_version_greater_than_dep(v, dep_version)]
        if fixed:
            closest_min_fix = min(fixed, key=version_sort_key)
            min_upgrade_type = get_upgrade_type(dep_version, closest_min_fix)
            min_residual_cves = _get_residual_cves(
                closest_min_fix, dep_version, advisories, base_advisories, base_keys
            )
            return FixMetadata(
                fix_version=closest_min_fix,
                upgrade_type=min_upgrade_type,
                breaking_change=min_upgrade_type == UpgradeType.MAJOR,
                residual_cves=min_residual_cves,
            )

    return None


def _calculate_complete_fix(  # noqa: PLR0913
    dep_version: str,
    advisories: list[Advisory],
    complete_fix_versions: list[str] | None,
    base_advisories: list[Advisory],
    base_keys: set[tuple[str, str]],
    ecosystem_str: str = "",
) -> FixMetadata | None:
    if complete_fix_versions:
        stable = _filter_stable_versions(complete_fix_versions, ecosystem_str)
        complete_fix_list = [v for v in stable if _is_version_greater_than_dep(v, dep_version)]
        if complete_fix_list:
            closest_complete_fix = min(complete_fix_list, key=version_sort_key)
            complete_upgrade_type = get_upgrade_type(dep_version, closest_complete_fix)
            complete_residual_cves = _get_residual_cves(
                closest_complete_fix, dep_version, advisories, base_advisories, base_keys
            )
            return FixMetadata(
                fix_version=closest_complete_fix,
                upgrade_type=complete_upgrade_type,
                breaking_change=complete_upgrade_type == UpgradeType.MAJOR,
                residual_cves=complete_residual_cves,
            )
    return None


def _calculate_safe_fix(  # noqa: PLR0913
    dep_version: str,
    minimal_fix: FixMetadata,
    complete_fix: FixMetadata,
    all_package_versions: list[str],
    advisories: list[Advisory],
    advisory: Advisory,
    base_advisories: list[Advisory],
    base_keys: set[tuple[str, str]],
    ecosystem_str: str = "",
) -> FixMetadata | None:
    minimal_fix_version = minimal_fix.fix_version
    complete_fix_version = complete_fix.fix_version
    stable_versions = _filter_stable_versions(all_package_versions, ecosystem_str)
    safe_version_candidates = [
        v
        for v in stable_versions
        if _is_version_greater_than_dep(v, minimal_fix_version)
        and _is_version_greater_than_dep(complete_fix_version, v)
    ]
    if safe_version_candidates:
        valid_safe_candidates = {}
        for candidate in safe_version_candidates:
            if match_vulnerable_versions(candidate, advisory.vulnerable_version):
                continue

            residual_cves = _get_residual_cves(
                candidate, dep_version, advisories, base_advisories, base_keys
            )
            if (
                residual_cves is not None
                and residual_cves.introduced is not None
                and residual_cves.introduced.total > 0
            ):
                continue

            valid_safe_candidates[candidate] = residual_cves

        if valid_safe_candidates:
            closest_safe_version = min(valid_safe_candidates.keys(), key=version_sort_key)
            residual_cves = valid_safe_candidates[closest_safe_version]
            safe_upgrade_type = get_upgrade_type(dep_version, closest_safe_version)
            return FixMetadata(
                fix_version=closest_safe_version,
                upgrade_type=safe_upgrade_type,
                breaking_change=safe_upgrade_type == UpgradeType.MAJOR,
                residual_cves=residual_cves,
            )
    return None


def match_fixed_versions_new(
    dep_version: str,
    advisory: Advisory,
    complete_fix_versions: list[str] | None,
    all_package_versions: list[str],
    advisories: list[Advisory],
) -> Advisory:
    # Precalculate base_advisories once for all residual CVE calculations
    base_advisories = [
        a for a in advisories if match_vulnerable_versions(dep_version, a.vulnerable_version)
    ]
    base_keys = {_advisory_key(a) for a in base_advisories}

    cve_fix_versions = advisory.fixed_versions

    eco = advisory.ecosystem
    minimal_fix = _calculate_minimal_fix(
        dep_version, advisories, cve_fix_versions, base_advisories, base_keys, eco
    )
    complete_fix = _calculate_complete_fix(
        dep_version, advisories, complete_fix_versions, base_advisories, base_keys, eco
    )

    if not minimal_fix and complete_fix:
        minimal_fix = complete_fix

    safe_fix = None
    if minimal_fix:
        residual = minimal_fix.residual_cves
        has_no_introduced = (
            residual is None or residual.introduced is None or residual.introduced.total == 0
        )
        if has_no_introduced:
            safe_fix = minimal_fix

    if minimal_fix and complete_fix and not safe_fix:
        safe_fix = _calculate_safe_fix(
            dep_version,
            minimal_fix,
            complete_fix,
            all_package_versions,
            advisories,
            advisory,
            base_advisories,
            base_keys,
            eco,
        )
    if not minimal_fix and not complete_fix and not safe_fix:
        advisory.fix_metadata_new = None
        return advisory

    if minimal_fix is None:
        advisory.fix_metadata_new = None
        return advisory

    advisory.fix_metadata_new = AdvisoryFixMetadataNew(
        closest_min_fix=minimal_fix,
        closest_complete_fix=complete_fix,
        closest_safe_fix=safe_fix,
    )

    return advisory


def match_fixes(
    dep_version: str,
    advisory: Advisory,
    complete_fix_versions: list[str] | None,
    all_package_versions: list[str],
    advisories: list[Advisory],
) -> Advisory:
    match_fixed_versions(dep_version, advisory, complete_fix_versions)
    match_fixed_versions_new(
        dep_version, advisory, complete_fix_versions, all_package_versions, advisories
    )
    return advisory



import logging
from pathlib import Path
from typing import Unpack, cast

from pydantic import BaseModel, StrictBool

from labels.model.core import (
    OutputFormat,
    SbomConfig,
    ScanArgs,
    SourceType,
)
from labels.model.sources import AwsCredentials
from labels.utils.exceptions import InvalidConfigFileError
from labels.utils.yaml import load_yaml_mapping

LOGGER = logging.getLogger(__name__)


class RawOutputConfig(BaseModel):
    name: str
    format: OutputFormat


class RawDockerCredentials(BaseModel):
    username: str
    password: str


class RawPlatformOverrides(BaseModel):
    os_override: str | None = None
    arch_override: str | None = None


class RawAwsCredentials(BaseModel):
    external_id: str | None = None
    role: str | None = None
    region: str | None = None
    access_key_id: str | None = None
    secret_access_key: str | None = None
    session_token: str | None = None


class RawConfig(BaseModel):
    model_config = {"extra": "allow"}

    source: str
    source_type: SourceType
    output: RawOutputConfig
    include: list[str] = ["."]
    exclude: list[str] = []
    execution_id: str | None = None
    docker_credentials: RawDockerCredentials | None = None
    platform_overrides: RawPlatformOverrides | None = None
    aws_credentials: RawAwsCredentials | None = None
    debug: StrictBool = False
    include_package_metadata: StrictBool = True
    include_package_advisories: StrictBool = True


class RawSbomSubConfig(BaseModel):
    output: RawOutputConfig
    include: list[str] = ["."]
    exclude: list[str] = []
    include_package_metadata: StrictBool = True
    include_package_advisories: StrictBool = True


class RawStaticScanConfig(BaseModel):
    model_config = {"extra": "allow"}

    sbom: RawSbomSubConfig
    source_type: SourceType = SourceType.DIRECTORY
    working_dir: str = "./"
    source: str | None = None
    execution_id: str | None = None
    docker_credentials: RawDockerCredentials | None = None
    platform_overrides: RawPlatformOverrides | None = None
    aws_credentials: RawAwsCredentials | None = None
    debug: StrictBool = False


def build_labels_config_from_args(arg: str, **kwargs: Unpack[ScanArgs]) -> SbomConfig:
    return SbomConfig(
        source=arg,
        source_type=SourceType.from_string(kwargs["source"]),
        execution_id=None,
        output_format=OutputFormat.from_string(kwargs["format"]),
        output=kwargs["output"],
        docker_user=kwargs["docker_user"],
        docker_password=kwargs["docker_password"],
        os_override=kwargs["os"],
        arch_override=kwargs["arch"],
        aws_external_id=kwargs["aws_external_id"],
        aws_role=kwargs["aws_role"],
        aws_region=kwargs["aws_region"],
        feature_preview=kwargs["feature_preview"],
        debug=kwargs["debug"],
    )


def build_labels_config_from_file(
    config_file_path: str,
    *,
    static_scan: bool = False,
) -> SbomConfig:
    if Path(config_file_path).is_file():
        if not config_file_path.endswith((".yaml", ".yml")):
            error_msg = "The configuration file must be a YAML format"
            raise InvalidConfigFileError(config_file_path, error_msg)
        if static_scan:
            return load_config_from_static_scan(config_file_path)
        return load_config_from_file(config_file_path)

    raise InvalidConfigFileError(config_file_path)


def _load_aws_credentials(
    raw_aws: RawAwsCredentials | None,
) -> AwsCredentials | None:
    if raw_aws and raw_aws.access_key_id is not None and raw_aws.secret_access_key is not None:
        return AwsCredentials(
            access_key_id=raw_aws.access_key_id,
            secret_access_key=raw_aws.secret_access_key,
            session_token=raw_aws.session_token,
        )
    return None


def _build_raw_config(config_path: str) -> RawConfig:
    raw_data = load_yaml_mapping(config_path)
    config = RawConfig.model_validate(raw_data)

    extras = cast("dict[str, object] | None", config.model_extra)
    if extras:
        _log_unrecognized_keys(extras)
    return config


def _build_raw_static_scan_config(config_path: str) -> RawStaticScanConfig:
    raw_data = load_yaml_mapping(config_path)
    config = RawStaticScanConfig.model_validate(raw_data)

    extras = cast("dict[str, object] | None", config.model_extra)
    if extras:
        _log_unrecognized_keys(extras)
    return config


def _log_unrecognized_keys(extras: dict[str, object]) -> None:
    unrecognized_keys = ", ".join(sorted(extras))
    msg = (
        f"Some keys were not recognized: {unrecognized_keys}."
        " The analysis will be performed only using the supported keys"
        " and defaults."
    )
    LOGGER.warning(msg)


def load_config_from_file(config_path: str) -> SbomConfig:
    config = _build_raw_config(config_path)

    docker = config.docker_credentials
    aws = config.aws_credentials
    overrides = config.platform_overrides

    return SbomConfig(
        source=config.source,
        source_type=config.source_type,
        output_format=config.output.format,
        output=config.output.name,
        include=tuple(config.include),
        exclude=tuple(config.exclude),
        docker_user=docker.username if docker else None,
        docker_password=docker.password if docker else None,
        os_override=overrides.os_override if overrides else None,
        arch_override=overrides.arch_override if overrides else None,
        aws_external_id=aws.external_id if aws else None,
        aws_role=aws.role if aws else None,
        aws_region=aws.region if aws else None,
        aws_credentials=_load_aws_credentials(aws),
        execution_id=config.execution_id,
        debug=config.debug,
        include_package_metadata=config.include_package_metadata,
        include_package_advisories=config.include_package_advisories,
    )


def load_config_from_static_scan(config_path: str) -> SbomConfig:
    config = _build_raw_static_scan_config(config_path)
    sbom = config.sbom

    docker = config.docker_credentials
    aws = config.aws_credentials
    overrides = config.platform_overrides

    source = (
        config.working_dir
        if config.source_type == SourceType.DIRECTORY or config.source is None
        else config.source
    )

    return SbomConfig(
        source=source,
        source_type=config.source_type,
        output_format=sbom.output.format,
        output=sbom.output.name,
        include=tuple(sbom.include),
        exclude=tuple(sbom.exclude),
        docker_user=docker.username if docker else None,
        docker_password=docker.password if docker else None,
        os_override=overrides.os_override if overrides else None,
        arch_override=overrides.arch_override if overrides else None,
        aws_external_id=aws.external_id if aws else None,
        aws_role=aws.role if aws else None,
        aws_region=aws.region if aws else None,
        aws_credentials=_load_aws_credentials(aws),
        execution_id=config.execution_id,
        debug=config.debug,
        include_package_metadata=sbom.include_package_metadata,
        include_package_advisories=sbom.include_package_advisories,
    )
